"""IDL coding."""
from dataclasses import fields as dc_fields
from dataclasses import make_dataclass
from keyword import kwlist
from types import MappingProxyType
from typing import Mapping, Type, cast

from anchorpy3_core.idl import (
    IdlField,
    IdlType,
    IdlTypeArray,
    IdlTypeDefined,
    IdlTypeDefinition,
    IdlTypeDefinitionTyAlias,
    IdlTypeDefinitionTyEnum,
    IdlTypeDefinitionTyStruct,
    IdlTypeOption,
    IdlTypeSimple,
    IdlTypeVec,
)
from anchorpy.coder.idl_compat import get_defined_type_name
from borsh_construct import (
    F32,
    F64,
    I8,
    I16,
    I32,
    I64,
    I128,
    U8,
    U16,
    U32,
    U64,
    U128,
    Bool,
    Bytes,
    CStruct,
    Enum,
    Option,
    String,
    TupleStruct,
    Vec,
)
from construct import Construct
from pyheck import snake

from anchorpy.borsh_extension import BorshPubkey, _DataclassStruct
from anchorpy.idl import TypeDefs

FIELD_TYPE_MAP: Mapping[IdlTypeSimple, Construct] = MappingProxyType(
    {
        IdlTypeSimple.Bool: Bool,
        IdlTypeSimple.U8: U8,
        IdlTypeSimple.I8: I8,
        IdlTypeSimple.U16: U16,
        IdlTypeSimple.I16: I16,
        IdlTypeSimple.U32: U32,
        IdlTypeSimple.I32: I32,
        IdlTypeSimple.F32: F32,
        IdlTypeSimple.U64: U64,
        IdlTypeSimple.I64: I64,
        IdlTypeSimple.F64: F64,
        IdlTypeSimple.U128: U128,
        IdlTypeSimple.I128: I128,
        IdlTypeSimple.Bytes: Bytes,
        IdlTypeSimple.String: String,
        IdlTypeSimple.PublicKey: BorshPubkey,
    },
)


_enums_cache: dict[tuple[str, str], Enum] = {}


def _handle_enum_variants(
    idl_enum,  # Can be IdlTypeDefinitionTyEnum or dict
    types: TypeDefs,
    name: str,
) -> Enum:
    dict_key = (name, str(idl_enum))
    try:
        return _enums_cache[dict_key]
    except KeyError:
        result = _handle_enum_variants_no_cache(idl_enum, types, name)
        _enums_cache[dict_key] = result
        return result


def _handle_enum_variants_no_cache(
    idl_enum,  # Can be IdlTypeDefinitionTyEnum or dict
    types: TypeDefs,
    name: str,
) -> Enum:
    variants = []
    dclasses = {}
    # Get variants handling both object and dict formats
    enum_variants = idl_enum.variants if hasattr(idl_enum, 'variants') else idl_enum.get('variants', [])
    for variant in enum_variants:
        # Handle both object and dict variants
        if hasattr(variant, 'name'):
            variant_name = variant.name
            variant_fields = variant.fields
        elif isinstance(variant, dict):
            variant_name = variant.get('name')
            variant_fields = variant.get('fields')
        else:
            continue

        if variant_fields is None:
            variants.append(variant_name)
        else:
            # Get fields from variant_fields (can be object or dict)
            if hasattr(variant_fields, 'fields'):
                flds = variant_fields.fields
            elif isinstance(variant_fields, dict):
                flds = variant_fields.get('fields', [])
            else:
                flds = variant_fields if isinstance(variant_fields, list) else []
            # Check if fields are named (have 'name' property) or tuple-style
            if flds and (isinstance(flds[0], IdlField) or (isinstance(flds[0], dict) and 'name' in flds[0])):
                fields = []
                named_fields = cast(list[IdlField], flds)
                for fld in named_fields:
                    fields.append(_field_layout(fld, types))
                cstruct = CStruct(*fields)
                datacls = _idl_enum_fields_named_to_dataclass_type(
                    named_fields,
                    variant_name,
                )
                dclasses[variant_name] = datacls
                renamed = variant_name / cstruct
            else:
                fields = []
                unnamed_fields = cast(list[IdlType], flds)
                for type_ in unnamed_fields:
                    fields.append(_type_layout(type_, types))
                tuple_struct = TupleStruct(*fields)
                renamed = variant_name / tuple_struct
            variants.append(renamed)  # type: ignore
    enum_without_types = Enum(*variants, enum_name=name)
    if dclasses:
        for cname in enum_without_types.enum._sumtype_constructor_names:
            try:
                dclass = dclasses[cname]
            except KeyError:
                continue
            dclass_fields = dc_fields(dclass)
            constructr = getattr(enum_without_types.enum, cname)
            for constructor_field in constructr._sumtype_attribs:
                attrib = constructor_field[1]  # type: ignore
                fld_name = constructor_field[0]  # type: ignore
                dclass_field = [f for f in dclass_fields if f.name == fld_name][0]
                attrib.type = dclass_field.type  # type: ignore
    return enum_without_types


def _typedef_layout_without_field_name(
    typedef,  # Can be IdlTypeDefinition or dict
    types: TypeDefs,
) -> Construct:
    # Handle both object and dict formats
    if hasattr(typedef, 'ty'):
        typedef_type = typedef.ty
        name = typedef.name
    elif isinstance(typedef, dict):
        typedef_type = typedef.get('ty') or typedef.get('type')
        name = typedef.get('name')
    else:
        raise ValueError(f"Unknown typedef format: {type(typedef)}")
    # Check typedef_type (can be object or dict)
    is_struct = isinstance(typedef_type, IdlTypeDefinitionTyStruct) or (
        isinstance(typedef_type, dict) and typedef_type.get('kind') == 'struct'
    )
    is_enum = isinstance(typedef_type, IdlTypeDefinitionTyEnum) or (
        isinstance(typedef_type, dict) and typedef_type.get('kind') == 'enum'
    )
    is_alias = isinstance(typedef_type, IdlTypeDefinitionTyAlias) or (
        isinstance(typedef_type, dict) and typedef_type.get('kind') == 'alias'
    )

    if is_struct:
        fields = typedef_type.fields if hasattr(typedef_type, 'fields') else typedef_type.get('fields', [])
        field_layouts = [_field_layout(field, types) for field in fields]
        cstruct = CStruct(*field_layouts)
        datacls = _idl_typedef_ty_struct_to_dataclass_type(typedef_type, name)
        return _DataclassStruct(cstruct, datacls=datacls)
    elif is_enum:
        return _handle_enum_variants(typedef_type, types, name)
    elif is_alias:
        value = typedef_type.value if hasattr(typedef_type, 'value') else typedef_type.get('value')
        return _type_layout(value, types)

    unknown_type = typedef_type.kind if hasattr(typedef_type, 'kind') else typedef_type.get('kind', 'unknown')
    raise ValueError(f"Unknown type {unknown_type}")


def _typedef_layout(
    typedef: IdlTypeDefinition,
    types: list[IdlTypeDefinition],
    field_name: str,
) -> Construct:
    """Map an IDL typedef to a `Construct` object.

    Args:
        typedef: The IDL typedef object.
        types: IDL type definitions.
        field_name: The name of the field.

    Raises:
        ValueError: If an unknown type is passed.

    Returns:
        `Construct` object from `borsh-construct`.
    """
    return field_name / _typedef_layout_without_field_name(typedef, types)


def _type_layout(type_: IdlType, types: TypeDefs) -> Construct:
    if isinstance(type_, IdlTypeSimple):
        return FIELD_TYPE_MAP[type_]
    if isinstance(type_, IdlTypeVec):
        return Vec(_type_layout(type_.vec, types))
    elif isinstance(type_, IdlTypeOption):
        return Option(_type_layout(type_.option, types))
    elif isinstance(type_, IdlTypeDefined):
        # Support both old string format and new object format
        defined = get_defined_type_name(type_.defined)
        if not types:
            raise ValueError("User defined types not provided")
        # Handle both object and dict types
        filtered = []
        for t in types:
            type_name = t.name if hasattr(t, 'name') else t.get('name')
            if type_name == defined:
                filtered.append(t)
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        return _typedef_layout_without_field_name(filtered[0], types)
    elif isinstance(type_, IdlTypeArray):
        array_ty = type_.array[0]
        array_len = type_.array[1]
        inner_layout = _type_layout(array_ty, types)
        return inner_layout[array_len]
    raise ValueError(f"Type {type_} not implemented yet")


def _field_layout(field, types: TypeDefs) -> Construct:  # field can be IdlField or dict
    """Map IDL spec to `borsh-construct` types.

    Args:
        field: field object from the IDL.
        types: IDL type definitions.

    Raises:
        ValueError: If the user-defined types are not provided.
        ValueError: If the type is not found.
        ValueError: If the type is not implemented yet.

    Returns:
        `Construct` object from `borsh-construct`.
    """
    # Handle both object and dict fields
    if hasattr(field, 'name'):
        field_name = snake(field.name) if field.name else ""
        field_ty = field.ty
    elif isinstance(field, dict):
        field_name = snake(field.get('name')) if field.get('name') else ""
        field_ty = field.get('ty') or field.get('type')
    else:
        raise ValueError(f"Unknown field format: {type(field)}")
    return field_name / _type_layout(field_ty, types)


def _make_datacls(name: str, fields: list[str]) -> type:
    return make_dataclass(name, fields)


_idl_typedef_ty_struct_to_dataclass_type_cache: dict[tuple[str, str], Type] = {}


def _idl_typedef_ty_struct_to_dataclass_type(
    typedef_type,  # Can be IdlTypeDefinitionTyStruct or dict
    name: str,
) -> Type:
    dict_key = (name, str(typedef_type))
    try:
        return _idl_typedef_ty_struct_to_dataclass_type_cache[dict_key]
    except KeyError:
        result = _idl_typedef_ty_struct_to_dataclass_type_no_cache(typedef_type, name)
        _idl_typedef_ty_struct_to_dataclass_type_cache[dict_key] = result
        return result


def _idl_typedef_ty_struct_to_dataclass_type_no_cache(
    typedef_type,  # Can be IdlTypeDefinitionTyStruct or dict
    name: str,
) -> Type:
    """Generate a dataclass definition from an IDL struct.

    Args:
        typedef_type: The IDL type.
        name: The name of the dataclass.

    Returns:
        Dataclass definition.
    """
    dataclass_fields = []
    # Get fields handling both object and dict formats
    fields = typedef_type.fields if hasattr(typedef_type, 'fields') else typedef_type.get('fields', [])

    for field in fields:
        # Handle both object and dict field formats
        if hasattr(field, 'name'):
            field_name = snake(field.name)
        elif isinstance(field, dict):
            field_name = snake(field.get('name', ''))
        else:
            continue

        field_name_to_use = f"{field_name}_" if field_name in kwlist else field_name
        dataclass_fields.append(
            field_name_to_use,
        )
    return _make_datacls(name, dataclass_fields)


_idl_enum_fields_named_to_dataclass_type_cache: dict[tuple[str, str], Type] = {}


def _idl_enum_fields_named_to_dataclass_type(
    fields: list[IdlField],
    name: str,
) -> Type:
    dict_key = (name, str(fields))
    try:
        return _idl_enum_fields_named_to_dataclass_type_cache[dict_key]
    except KeyError:
        result = _idl_enum_fields_named_to_dataclass_type_no_cache(fields, name)
        _idl_enum_fields_named_to_dataclass_type_cache[dict_key] = result
        return result


def _idl_enum_fields_named_to_dataclass_type_no_cache(
    fields: list[IdlField],
    name: str,
) -> Type:
    """Generate a dataclass definition from IDL named enum fields.

    Args:
        fields: The IDL enum fields.
        name: The name of the dataclass.

    Returns:
        Dataclass type definition.
    """
    dataclass_fields = []
    for field in fields:
        field_name = snake(field.name)
        field_name_to_use = f"{field_name}_" if field_name in kwlist else field_name
        dataclass_fields.append(
            field_name_to_use,
        )
    return _make_datacls(name, dataclass_fields)


def _idl_typedef_to_python_type(
    typedef: IdlTypeDefinition,
    types: TypeDefs,
) -> Type:
    """Generate Python type from IDL user-defined type.

    Args:
        typedef: The user-defined type.
        types: IDL type definitions.

    Raises:
        ValueError: If an unknown type is passed.

    Returns:
        The Python type.
    """
    typedef_type = typedef.ty
    if isinstance(typedef_type, IdlTypeDefinitionTyStruct):
        return _idl_typedef_ty_struct_to_dataclass_type(
            typedef_type,
            typedef.name,
        )
    elif isinstance(typedef_type, IdlTypeDefinitionTyEnum):
        return _handle_enum_variants(typedef_type, types, typedef.name).enum
    elif isinstance(typedef_type, IdlTypeDefinitionTyAlias):
        raise ValueError(f"Alias not handled here: {typedef_type}")
    unknown_type = typedef_type.kind
    raise ValueError(f"Unknown type {unknown_type}")



import typing
from anchorpy.error import ProgramError


class SomeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Example error.")

    code = 6000
    name = "SomeError"
    msg = "Example error."


class OtherError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Another error.")

    code = 6001
    name = "OtherError"
    msg = "Another error."


CustomError = typing.Union[SomeError, OtherError]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: SomeError(),
    6001: OtherError(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



"""This module contains code for creating the Anchor workspace."""
from pathlib import Path
from typing import Dict, Optional, Union

import toml  # type: ignore
from anchorpy3_core.idl import Idl
from solders.pubkey import Pubkey

from anchorpy.program.core import Program
from anchorpy.provider import Provider

WorkspaceType = Dict[str, Program]


def create_workspace(
    path: Optional[Union[Path, str]] = None, url: Optional[str] = None
) -> WorkspaceType:
    """Get a workspace from the provided path to the project root.

    Args:
        path: The path to the project root. Defaults to the current working
            directory if omitted.
        url: The URL of the JSON RPC. Defaults to http://localhost:8899.

    Returns:
        Mapping of program name to Program object.
    """
    result = {}
    project_root = Path.cwd() if path is None else Path(path)
    idl_folder = project_root / "target/idl"
    localnet_programs: dict[str, str] = toml.load(project_root / "Anchor.toml")[
        "programs"
    ]["localnet"]
    for file in idl_folder.iterdir():
        raw = file.read_text()
        idl = Idl.from_json(raw)
        name = idl.name
        program_id = Pubkey.from_string(localnet_programs[name])
        program = Program(idl, program_id, Provider.local(url))
        result[idl.name] = program
    return result


async def close_workspace(workspace: WorkspaceType) -> None:
    """Close the HTTP clients of all the programs in the workspace.

    Args:
        workspace: The workspace to close.
    """
    for program in workspace.values():
        # could do this in a faster way but there's probably no point.
        await program.close()



"""This module deals with (de)serializing Anchor events."""
from hashlib import sha256
from typing import Any, Dict, Optional, Tuple

from anchorpy3_core.idl import (
    Idl,
    IdlEvent,
    IdlField,
    IdlTypeDefinition,
    IdlTypeDefinitionTyStruct,
)
from construct import Adapter, Bytes, Construct, Sequence, Switch
from pyheck import snake

from anchorpy.coder.idl import _typedef_layout, _typedef_layout_without_field_name
from anchorpy.coder.idl_compat import get_event_discriminator
from anchorpy.program.common import Event


def _event_discriminator(name: str) -> bytes:
    """Get 8-byte discriminator from event name.

    Args:
        name: The event name.

    Returns:
        Discriminator
    """
    return sha256(f"event:{name}".encode()).digest()[:8]


def _event_layout(event, idl: Idl) -> Construct:  # event can be IdlEvent or dict
    # Handle both object and dict events
    if hasattr(event, 'name'):
        event_name = event.name
        event_fields = getattr(event, 'fields', None)
    elif isinstance(event, dict):
        event_name = event.get('name')
        event_fields = event.get('fields')
    else:
        raise ValueError(f"Unknown event format: {type(event)}")

    # For new format, events might not have fields directly
    # Need to look up in types array
    if event_fields is None:
        # Try to find the event type in the types array
        for type_def in idl.types:
            type_name = type_def.name if hasattr(type_def, 'name') else type_def.get('name')
            if type_name == event_name:
                return _typedef_layout_without_field_name(type_def, idl.types)

        # If not found in types, create empty struct
        event_type_def = IdlTypeDefinition(
            name=event_name,
            docs=None,
            ty=IdlTypeDefinitionTyStruct(fields=[]),
        )
    else:
        # Old format with fields
        event_type_def = IdlTypeDefinition(
            name=event_name,
            docs=None,
            ty=IdlTypeDefinitionTyStruct(
                fields=[
                    IdlField(name=snake(f.name), docs=None, ty=f.ty) for f in event_fields
                ],
            ),
        )

    return _typedef_layout(event_type_def, idl.types, event_name)


class EventCoder(Adapter):
    """Encodes and decodes Anchor events."""

    def __init__(self, idl: Idl):
        """Initialize the EventCoder.

        Args:
            idl: The parsed Idl object.
        """
        self.idl = idl
        idl_events = idl.events
        layouts: Dict[str, Construct]
        if idl_events:
            layouts = {}
            for event in idl_events:
                # Handle both object and dict events
                if hasattr(event, 'name'):
                    event_name = event.name
                elif isinstance(event, dict):
                    event_name = event.get('name')
                else:
                    continue
                layouts[event_name] = _event_layout(event, idl)
        else:
            layouts = {}
        self.layouts = layouts
        # Support both calculated and precomputed discriminators
        self.discriminators: Dict[bytes, str] = {}
        if idl_events:
            for event in idl_events:
                # Get event name (handle both object and dict)
                if hasattr(event, 'name'):
                    event_name = event.name
                elif isinstance(event, dict):
                    event_name = event.get('name')
                else:
                    continue

                # New format: use precomputed discriminator if available
                disc_list = get_event_discriminator(event)
                if disc_list:
                    disc = bytes(disc_list)
                else:
                    # Old format: calculate from name
                    disc = _event_discriminator(event_name)
                self.discriminators[disc] = event_name
        self.discriminator_to_layout = {
            disc: self.layouts[event_name]
            for disc, event_name in self.discriminators.items()
        }
        subcon = Sequence(
            "discriminator" / Bytes(8),  # not base64-encoded here
            Switch(lambda this: this.discriminator, self.discriminator_to_layout),
        )
        super().__init__(subcon)  # type: ignore

    def _decode(self, obj: Tuple[bytes, Any], context, path) -> Optional[Event]:
        disc = obj[0]
        try:
            event_name = self.discriminators[disc]
        except KeyError:
            return None
        return Event(data=obj[1], name=event_name)



#!/usr/bin/env python3
"""Comprehensive test suite for new and old IDL format support in AnchorPy.

This test suite verifies that AnchorPy correctly handles both:
- Old (legacy) Anchor IDL format
- New (v0.1.0 spec) Anchor IDL format

Run with: pytest tests/test_idl_format_suite.py -v
"""

import json
import tempfile
from pathlib import Path
import pytest
from anchorpy3_core.idl import Idl
from anchorpy import Program, Provider
from anchorpy.clientgen.accounts import gen_accounts
from anchorpy.clientgen.errors import gen_errors
from anchorpy.clientgen.instructions import gen_instructions
from anchorpy.clientgen.program_id import gen_program_id
from anchorpy.clientgen.types import gen_types


class TestIDLFormatSupport:
    """Test suite for IDL format support."""

    @pytest.fixture
    def provider(self):
        """Create a provider for testing."""
        return Provider.local()

    def get_test_idl_path(self, filename: str) -> Path:
        """Get the path to a test IDL file."""
        return Path(__file__).parent.parent / filename

    @pytest.mark.parametrize("idl_file,format_type", [
        ("kamino_lend_v4.json", "old"),
        ("adrena.json", "old"),
        ("loopscale_v1.json", "new"),
        ("loopscale_v2.json", "new"),
    ])
    def test_idl_parsing(self, idl_file: str, format_type: str):
        """Test that IDL files can be parsed correctly."""
        idl_path = self.get_test_idl_path(idl_file)

        with open(idl_path) as f:
            idl_json = json.load(f)

        # Parse as Idl object
        idl_obj = Idl.from_json(json.dumps(idl_json))

        assert idl_obj is not None
        assert idl_obj.instructions is not None

        # Check format detection
        if format_type == "new":
            # New format should have discriminators
            if idl_obj.accounts and len(idl_obj.accounts) > 0:
                first_acc = idl_obj.accounts[0]
                # Check for discriminator (it may be present in new format)
                assert hasattr(first_acc, 'name')

    @pytest.mark.parametrize("idl_file", [
        "kamino_lend_v4.json",
        "adrena.json",
        "loopscale_v1.json",
        "loopscale_v2.json",
    ])
    def test_program_creation(self, idl_file: str, provider):
        """Test that Program objects can be created from IDL files."""
        idl_path = self.get_test_idl_path(idl_file)

        with open(idl_path) as f:
            idl_json = json.load(f)

        # Get program ID
        program_id = idl_json.get('address', 'So11111111111111111111111111111111111111112')

        # Create Program object
        program = Program(idl_json, program_id, provider)

        assert program is not None
        assert program.program_id == program_id
        assert hasattr(program, 'rpc')
        assert hasattr(program, 'account')

    @pytest.mark.parametrize("idl_file", [
        "kamino_lend_v4.json",
        "adrena.json",
        "loopscale_v1.json",
        "loopscale_v2.json",
    ])
    def test_client_generation(self, idl_file: str):
        """Test that client code can be generated from IDL files."""
        idl_path = self.get_test_idl_path(idl_file)

        with open(idl_path) as f:
            idl_json = json.load(f)

        # Parse as Idl object
        idl_obj = Idl.from_json(json.dumps(idl_json))

        # Get program ID
        program_id = idl_json.get('address')
        if not program_id:
            metadata = idl_json.get('metadata', {})
            program_id = metadata.get('address')
        if not program_id:
            program_id = 'So11111111111111111111111111111111111111112'

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = Path(tmpdir) / "generated"
            output_path.mkdir()

            # Generate all client code
            gen_program_id(program_id, output_path)
            gen_errors(idl_obj, output_path)
            gen_instructions(idl_obj, output_path, gen_pdas=False)
            gen_types(idl_obj, output_path)
            gen_accounts(idl_obj, output_path)

            # Verify files were created
            assert (output_path / "program_id.py").exists()

            # Check for generated content
            if idl_obj.accounts:
                account_files = list((output_path / "accounts").glob("*.py"))
                assert len(account_files) > 0

            if idl_obj.instructions:
                instruction_files = list((output_path / "instructions").glob("*.py"))
                assert len(instruction_files) > 0

    def test_discriminator_handling(self):
        """Test that discriminators are handled correctly in new format."""
        idl_path = self.get_test_idl_path("loopscale_v1.json")

        with open(idl_path) as f:
            idl_json = json.load(f)

        idl_obj = Idl.from_json(json.dumps(idl_json))

        # Check that accounts have discriminators in new format
        if idl_obj.accounts:
            for account in idl_obj.accounts:
                # In new format, discriminators should be precomputed
                if hasattr(account, 'discriminator'):
                    assert account.discriminator is not None
                    assert isinstance(account.discriminator, list)
                    assert len(account.discriminator) == 8

    def test_backward_compatibility(self, provider):
        """Test that old format IDLs still work correctly."""
        idl_path = self.get_test_idl_path("kamino_lend_v4.json")

        with open(idl_path) as f:
            idl_json = json.load(f)

        # Should work exactly as before
        program = Program(
            idl_json,
            idl_json.get('address', 'So11111111111111111111111111111111111111112'),
            provider
        )

        assert program is not None
        assert hasattr(program, 'rpc')
        assert hasattr(program, 'account')


if __name__ == "__main__":
    # Run tests with pytest
    pytest.main([__file__, "-v"])


from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class CreateArgs(typing.TypedDict):
    authority: Pubkey


layout = borsh.CStruct("authority" / BorshPubkey)


class CreateAccounts(typing.TypedDict):
    counter: Pubkey


def create(
    args: CreateArgs,
    accounts: CreateAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["counter"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x18\x1e\xc8(\x05\x1c\x07w"
    encoded_args = layout.build(
        {
            "authority": args["authority"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



"""The TypesCoder class is for encoding and decoding user-defined types."""

from typing import Any, Dict

from anchorpy3_core.idl import Idl
from construct import Construct, Container

from anchorpy.coder.idl import _typedef_layout_without_field_name


class TypesCoder:
    """Encodes and decodes user-defined types in Anchor programs."""

    def __init__(self, idl: Idl) -> None:
        """Initialize the TypesCoder.

        Args:
            idl: The parsed IDL object.
        """
        self.idl = idl
        self.types_layouts: Dict[str, Construct] = {}

        self.filtered_types = []
        if idl.types:
            self.filtered_types = [
                ty for ty in idl.types if not getattr(ty, "generics", None)
            ]

    def _get_layout(self, name: str) -> Construct:
        """Get or create a layout for a given type name.

        Args:
            name: The name of the type.

        Returns:
            The construct layout for the type.

        Raises:
            ValueError: If the type is not found.
        """
        if name in self.types_layouts:
            return self.types_layouts[name]

        type_defs = [t for t in self.filtered_types if t.name == name]
        if not type_defs:
            raise ValueError(f"Unknown type: {name}")

        layout = _typedef_layout_without_field_name(type_defs[0], self.idl.types)
        self.types_layouts[name] = layout
        return layout

    def encode(self, name: str, data: Any) -> bytes:
        """Encode a user-defined type.

        Args:
            name: The name of the type.
            data: The data to encode.

        Returns:
            The encoded data.

        Raises:
            ValueError: If the type is not found.
        """
        layout = self._get_layout(name)
        return layout.build(data)

    def decode(self, name: str, buffer: bytes) -> Container[Any]:
        """Decode a user-defined type.

        Args:
            name: The name of the type.
            buffer: The buffer to decode.

        Returns:
            The decoded data.

        Raises:
            ValueError: If the type is not found.
        """
        layout = self._get_layout(name)
        return layout.parse(buffer)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class InitializedJSON(typing.TypedDict):
    kind: typing.Literal["Initialized"]


class FrozenJSON(typing.TypedDict):
    kind: typing.Literal["Frozen"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Initialized:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Initialized"

    @classmethod
    def to_json(cls) -> InitializedJSON:
        return InitializedJSON(
            kind="Initialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Initialized": {},
        }


@dataclass
class Frozen:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Frozen"

    @classmethod
    def to_json(cls) -> FrozenJSON:
        return FrozenJSON(
            kind="Frozen",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Frozen": {},
        }


AccountStateKind = typing.Union[Uninitialized, Initialized, Frozen]
AccountStateJSON = typing.Union[UninitializedJSON, InitializedJSON, FrozenJSON]


def from_decoded(obj: dict) -> AccountStateKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Initialized" in obj:
        return Initialized()
    if "Frozen" in obj:
        return Frozen()
    raise ValueError("Invalid enum object")


def from_json(obj: AccountStateJSON) -> AccountStateKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Initialized":
        return Initialized()
    if obj["kind"] == "Frozen":
        return Frozen()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "Initialized" / borsh.CStruct(),
    "Frozen" / borsh.CStruct(),
)



from .initialize_mint import initialize_mint, InitializeMintArgs, InitializeMintAccounts
from .initialize_account import initialize_account, InitializeAccountAccounts
from .initialize_multisig import (
    initialize_multisig,
    InitializeMultisigArgs,
    InitializeMultisigAccounts,
)
from .transfer import transfer, TransferArgs, TransferAccounts
from .approve import approve, ApproveArgs, ApproveAccounts
from .revoke import revoke, RevokeAccounts
from .set_authority import set_authority, SetAuthorityArgs, SetAuthorityAccounts
from .mint_to import mint_to, MintToArgs, MintToAccounts
from .burn import burn, BurnArgs, BurnAccounts
from .close_account import close_account, CloseAccountAccounts
from .freeze_account import freeze_account, FreezeAccountAccounts
from .thaw_account import thaw_account, ThawAccountAccounts
from .transfer_checked import (
    transfer_checked,
    TransferCheckedArgs,
    TransferCheckedAccounts,
)
from .approve_checked import approve_checked, ApproveCheckedArgs, ApproveCheckedAccounts
from .mint_to_checked import mint_to_checked, MintToCheckedArgs, MintToCheckedAccounts
from .burn_checked import burn_checked, BurnCheckedArgs, BurnCheckedAccounts
from .initialize_account2 import (
    initialize_account2,
    InitializeAccount2Args,
    InitializeAccount2Accounts,
)
from .sync_native import sync_native, SyncNativeAccounts
from .initialize_account3 import (
    initialize_account3,
    InitializeAccount3Args,
    InitializeAccount3Accounts,
)
from .initialize_multisig2 import (
    initialize_multisig2,
    InitializeMultisig2Args,
    InitializeMultisig2Accounts,
)
from .initialize_mint2 import (
    initialize_mint2,
    InitializeMint2Args,
    InitializeMint2Accounts,
)
from .get_account_data_size import get_account_data_size, GetAccountDataSizeAccounts
from .initialize_immutable_owner import (
    initialize_immutable_owner,
    InitializeImmutableOwnerAccounts,
)
from .amount_to_ui_amount import (
    amount_to_ui_amount,
    AmountToUiAmountArgs,
    AmountToUiAmountAccounts,
)
from .ui_amount_to_amount import (
    ui_amount_to_amount,
    UiAmountToAmountArgs,
    UiAmountToAmountAccounts,
)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey, COption
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class InitializeMintArgs(typing.TypedDict):
    decimals: int
    mint_authority: Pubkey
    freeze_authority: typing.Optional[Pubkey]


layout = borsh.CStruct(
    "decimals" / borsh.U8,
    "mint_authority" / BorshPubkey,
    "freeze_authority" / COption(BorshPubkey),
)


class InitializeMintAccounts(typing.TypedDict):
    mint: Pubkey


def initialize_mint(
    args: InitializeMintArgs,
    accounts: InitializeMintAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xd1*\xc3\x04\x81U\xd1,"
    encoded_args = layout.build(
        {
            "decimals": args["decimals"],
            "mint_authority": args["mint_authority"],
            "freeze_authority": args["freeze_authority"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TileJSON(typing.TypedDict):
    row: int
    column: int


@dataclass
class Tile:
    layout: typing.ClassVar = borsh.CStruct("row" / borsh.U8, "column" / borsh.U8)
    row: int
    column: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "Tile":
        return cls(row=obj.row, column=obj.column)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"row": self.row, "column": self.column}

    def to_json(self) -> TileJSON:
        return {"row": self.row, "column": self.column}

    @classmethod
    def from_json(cls, obj: TileJSON) -> "Tile":
        return cls(row=obj["row"], column=obj["column"])



from dataclasses import dataclass
from typing import Any, List, Optional

from solana.rpc import types
from solders.hash import Hash
from solders.instruction import AccountMeta, Instruction
from solders.keypair import Keypair
from solders.signature import Signature
from solders.transaction import VersionedTransaction

from anchorpy.program.context import Accounts, Context
from anchorpy.program.namespace.instruction import _InstructionFn
from anchorpy.program.namespace.rpc import _RpcFn
from anchorpy.program.namespace.simulate import SimulateResponse, _SimulateFn
from anchorpy.program.namespace.transaction import _TransactionFn


@dataclass
class IdlFuncs:
    ix_fn: _InstructionFn
    tx_fn: _TransactionFn
    rpc_fn: _RpcFn
    simulate_fn: _SimulateFn


class MethodsBuilder:
    def __init__(
        self,
        idl_funcs: IdlFuncs,
        accounts: Accounts,
        remaining_accounts: List[AccountMeta],
        signers: List[Keypair],
        pre_instructions: List[Instruction],
        post_instructions: List[Instruction],
        args: List[Any],
    ) -> None:
        self._idl_funcs = idl_funcs
        self._accounts = accounts
        self._remaining_accounts = remaining_accounts
        self._signers = signers
        self._pre_instructions = pre_instructions
        self._post_instructions = post_instructions
        self._args = args

    async def rpc(self, opts: Optional[types.TxOpts] = None) -> Signature:
        ctx = self._build_context(opts)
        return await self._idl_funcs.rpc_fn(*self._args, ctx=ctx)

    async def simulate(self, opts: Optional[types.TxOpts] = None) -> SimulateResponse:
        ctx = self._build_context(opts)
        return await self._idl_funcs.simulate_fn(*self._args, ctx=ctx)

    def instruction(self) -> Instruction:
        ctx = self._build_context(opts=None)
        return self._idl_funcs.ix_fn(*self._args, ctx=ctx)

    def transaction(self, payer: Keypair, blockhash: Hash) -> VersionedTransaction:
        ctx = self._build_context(opts=None)
        return self._idl_funcs.tx_fn(
            *self._args, ctx=ctx, payer=payer, blockhash=blockhash
        )

    def pubkeys(self) -> Accounts:
        return self._accounts

    def args(self, arguments: List[Any]) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions,
            args=arguments,
        )

    def accounts(self, accs: Accounts) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=accs,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions,
            args=self._args,
        )

    def signers(self, signers: List[Keypair]) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers + signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions,
            args=self._args,
        )

    def remaining_accounts(self, accounts: List[AccountMeta]) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts + accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions,
            args=self._args,
        )

    def pre_instructions(self, ixs: List[Instruction]) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions + ixs,
            post_instructions=self._post_instructions,
            args=self._args,
        )

    def post_instructions(self, ixs: List[Instruction]) -> "MethodsBuilder":
        idl_funcs = self._idl_funcs
        return MethodsBuilder(
            idl_funcs=idl_funcs,
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions + ixs,
            args=self._args,
        )

    def _build_context(self, opts: Optional[types.TxOpts]) -> Context:
        return Context(
            accounts=self._accounts,
            remaining_accounts=self._remaining_accounts,
            signers=self._signers,
            pre_instructions=self._pre_instructions,
            post_instructions=self._post_instructions,
            options=opts,
        )


def _build_methods_item(idl_funcs: IdlFuncs) -> MethodsBuilder:
    return MethodsBuilder(
        idl_funcs=idl_funcs,
        accounts={},
        remaining_accounts=[],
        signers=[],
        pre_instructions=[],
        post_instructions=[],
        args=[],
    )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class BurnCheckedArgs(typing.TypedDict):
    amount: int
    decimals: int


layout = borsh.CStruct("amount" / borsh.U64, "decimals" / borsh.U8)


class BurnCheckedAccounts(typing.TypedDict):
    account: Pubkey
    mint: Pubkey
    authority: Pubkey


def burn_checked(
    args: BurnCheckedArgs,
    accounts: BurnCheckedAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xc6y\xc8fx\xd0\x9b\xb2"
    encoded_args = layout.build(
        {
            "amount": args["amount"],
            "decimals": args["decimals"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class RevokeAccounts(typing.TypedDict):
    source: Pubkey
    owner: Pubkey


def revoke(
    accounts: RevokeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["source"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b'\xaa\x17\x1f"\x85\xad]\xf2'
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



"""Common utilities for encoding and decoding."""
from hashlib import sha256
from typing import Dict, Union

from anchorpy3_core.idl import (
    Idl,
    IdlEnumVariant,
    IdlField,
    IdlType,
    IdlTypeArray,
    IdlTypeCompound,
    IdlTypeDefined,
    IdlTypeDefinition,
    IdlTypeDefinitionTyAlias,
    IdlTypeDefinitionTyEnum,
    IdlTypeOption,
    IdlTypeSimple,
    IdlTypeVec,
)


def _sighash(ix_name: str) -> bytes:
    """Not technically sighash, since we don't include the arguments.

    (Because Rust doesn't allow function overloading.)

    Args:
        ix_name: The instruction name.

    Returns:
        The sighash bytes.
    """
    formatted_str = f"global:{ix_name}"
    return sha256(formatted_str.encode()).digest()[:8]


def _type_size_compound_type(idl: Idl, ty: IdlTypeCompound) -> int:
    if isinstance(ty, IdlTypeVec):
        return 1
    if isinstance(ty, IdlTypeOption):
        return 1 + _type_size(idl, ty.option)
    if isinstance(ty, IdlTypeDefined):
        defined = ty.defined
        filtered = [t for t in idl.types if t.name == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {ty}")
        type_def = filtered[0]
        return _account_size(idl, type_def)
    if isinstance(ty, IdlTypeArray):
        element_type = ty.array[0]
        array_size = ty.array[1]
        return _type_size(idl, element_type) * array_size
    raise ValueError(f"type_size not implemented for {ty}")


def _type_size(idl: Idl, ty: IdlType) -> int:
    """Return the size of the type in bytes.

    For variable length types, just return 1.
    Users should override this value in such cases.

    Args:
        idl: The parsed `Idl` object.
        ty: The type object from the IDL.

    Returns:
        The size of the object in bytes.
    """
    sizes: Dict[IdlTypeSimple, int] = {
        IdlTypeSimple.Bool: 1,
        IdlTypeSimple.U8: 1,
        IdlTypeSimple.I8: 1,
        IdlTypeSimple.Bytes: 1,
        IdlTypeSimple.String: 1,
        IdlTypeSimple.I16: 2,
        IdlTypeSimple.U16: 2,
        IdlTypeSimple.U32: 4,
        IdlTypeSimple.I32: 4,
        IdlTypeSimple.F32: 4,
        IdlTypeSimple.U64: 8,
        IdlTypeSimple.I64: 8,
        IdlTypeSimple.F64: 8,
        IdlTypeSimple.U128: 16,
        IdlTypeSimple.I128: 16,
        IdlTypeSimple.PublicKey: 32,
    }
    if isinstance(ty, IdlTypeSimple):
        return sizes[ty]
    return _type_size_compound_type(idl, ty)


def _variant_field_size(idl: Idl, field: Union[IdlField, IdlType]) -> int:
    if isinstance(field, IdlField):
        return _type_size(idl, field.ty)
    return _type_size(idl, field)


def _variant_size(idl: Idl, variant: IdlEnumVariant) -> int:
    if variant.fields is None:
        return 0
    field_sizes = []
    field: Union[IdlField, IdlType]
    for field in variant.fields.fields:
        field_sizes.append(_variant_field_size(idl, field))
    return sum(field_sizes)


def _account_size(idl: Idl, idl_account) -> int:
    """Calculate account size in bytes.

    Args:
        idl: The parsed `Idl` instance.
        idl_account: An item from `idl.accounts`.

    Returns:
        Account size.
    """
    # Handle both object and dict formats
    if hasattr(idl_account, 'ty'):
        idl_account_type = idl_account.ty
    elif isinstance(idl_account, dict):
        # For dict format, check both 'ty' and 'type' keys
        if 'ty' in idl_account:
            idl_account_type = idl_account['ty']
        elif 'type' in idl_account:
            idl_account_type = idl_account['type']
        else:
            # If neither exists, treat the whole dict as the type
            idl_account_type = idl_account
    else:
        idl_account_type = idl_account
    # Handle both object and dict formats for type checking
    if isinstance(idl_account_type, IdlTypeDefinitionTyEnum):
        variant_sizes = (
            _variant_size(idl, variant) for variant in idl_account_type.variants
        )
        return max(variant_sizes) + 1
    elif isinstance(idl_account_type, dict) and idl_account_type.get('kind') == 'enum':
        # Dict format enum
        variant_sizes = (
            _variant_size(idl, variant) for variant in idl_account_type.get('variants', [])
        )
        return max(variant_sizes) + 1

    if isinstance(idl_account_type, IdlTypeDefinitionTyAlias):
        return _type_size(idl, idl_account_type.value)
    elif isinstance(idl_account_type, dict) and idl_account_type.get('kind') == 'alias':
        # Dict format alias
        return _type_size(idl, idl_account_type.get('value'))

    # Handle struct fields
    fields = None
    if hasattr(idl_account_type, 'fields'):
        fields = idl_account_type.fields
    elif isinstance(idl_account_type, dict) and 'fields' in idl_account_type:
        fields = idl_account_type['fields']

    if fields is None:
        return 0

    # Calculate field sizes
    if isinstance(fields, list) and fields:
        if hasattr(fields[0], 'ty'):
            # Object format fields
            return sum(_type_size(idl, f.ty) for f in fields)
        elif isinstance(fields[0], dict) and 'type' in fields[0]:
            # Dict format fields
            return sum(_type_size(idl, f['type']) for f in fields)
        else:
            # Assume simple type list (tuple struct)
            return sum(_type_size(idl, f) for f in fields)
    return 0



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class XJSON(typing.TypedDict):
    kind: typing.Literal["X"]


class OJSON(typing.TypedDict):
    kind: typing.Literal["O"]


@dataclass
class X:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "X"

    @classmethod
    def to_json(cls) -> XJSON:
        return XJSON(
            kind="X",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "X": {},
        }


@dataclass
class O:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "O"

    @classmethod
    def to_json(cls) -> OJSON:
        return OJSON(
            kind="O",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "O": {},
        }


SignKind = typing.Union[X, O]
SignJSON = typing.Union[XJSON, OJSON]


def from_decoded(obj: dict) -> SignKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "X" in obj:
        return X()
    if "O" in obj:
        return O()
    raise ValueError("Invalid enum object")


def from_json(obj: SignJSON) -> SignKind:
    if obj["kind"] == "X":
        return X()
    if obj["kind"] == "O":
        return O()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("X" / borsh.CStruct(), "O" / borsh.CStruct())



INIT_TESTS = '''"""AnchorPy integration tests."""
import asyncio
from pytest import fixture, mark

from anchorpy import Program
from anchorpy.pytest_plugin import workspace_fixture
from anchorpy.workspace import WorkspaceType


@fixture(scope="module")
def event_loop():
    """Create a module-scoped event loop so we can use module-scope async fixtures."""
    loop = asyncio.get_event_loop_policy().new_event_loop()  # noqa: DAR301
    yield loop
    loop.close()


workspace = workspace_fixture(".")


@fixture(scope="module")
def program(workspace: WorkspaceType) -> Program:
    """Create a Program instance."""
    return workspace["{}"]


@mark.asyncio
async def test_init(program: Program) -> None:
    """Test that the initialize function is invoked successfully."""
    await program.rpc["initialize"]()
'''



import typing
import re
from solders.transaction_status import (
    InstructionErrorCustom,
    TransactionErrorInstructionError,
)
from solana.rpc.core import RPCException
from solders.rpc.errors import SendTransactionPreflightFailureMessage
from anchorpy.error import extract_code_and_logs
from ..program_id import PROGRAM_ID
from . import anchor
from . import custom


def from_code(code: int) -> typing.Union[custom.CustomError, anchor.AnchorError, None]:
    return custom.from_code(code) if code >= 6000 else anchor.from_code(code)


error_re = re.compile(r"Program (\w+) failed: custom program error: (\w+)")


def from_tx_error(
    error: RPCException,
) -> typing.Union[anchor.AnchorError, custom.CustomError, None]:
    err_info = error.args[0]
    extracted = extract_code_and_logs(err_info, PROGRAM_ID)
    if extracted is None:
        return None
    return from_code(extracted[0])



import typing
from anchorpy.error import ProgramError


class TileOutOfBounds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, None)

    code = 6000
    name = "TileOutOfBounds"
    msg = None


class TileAlreadySet(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, None)

    code = 6001
    name = "TileAlreadySet"
    msg = None


class GameAlreadyOver(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, None)

    code = 6002
    name = "GameAlreadyOver"
    msg = None


class NotPlayersTurn(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, None)

    code = 6003
    name = "NotPlayersTurn"
    msg = None


class GameAlreadyStarted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, None)

    code = 6004
    name = "GameAlreadyStarted"
    msg = None


CustomError = typing.Union[
    TileOutOfBounds, TileAlreadySet, GameAlreadyOver, NotPlayersTurn, GameAlreadyStarted
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: TileOutOfBounds(),
    6001: TileAlreadySet(),
    6002: GameAlreadyOver(),
    6003: NotPlayersTurn(),
    6004: GameAlreadyStarted(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey, COption
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class SetAuthorityArgs(typing.TypedDict):
    authority_type: types.authority_type.AuthorityTypeKind
    new_authority: typing.Optional[Pubkey]


layout = borsh.CStruct(
    "authority_type" / types.authority_type.layout,
    "new_authority" / COption(BorshPubkey),
)


class SetAuthorityAccounts(typing.TypedDict):
    owned: Pubkey
    owner: Pubkey
    signer: Pubkey


def set_authority(
    args: SetAuthorityArgs,
    accounts: SetAuthorityAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["owned"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["signer"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x85\xfa%\x15n\xa3\x1ay"
    encoded_args = layout.build(
        {
            "authority_type": args["authority_type"].to_encodable(),
            "new_authority": args["new_authority"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



"""Tests for new Anchor IDL format (v0.1.0 spec) compatibility."""
import json
from pathlib import Path

import pytest
from anchorpy import Program, Provider
from anchorpy3_core.idl import Idl


@pytest.fixture
def loopscale_idl():
    """Load Loopscale v1 IDL (new format)."""
    idl_path = Path(__file__).parent.parent.parent / "loopscale_v1.json"
    with open(idl_path) as f:
        return json.load(f)


@pytest.fixture
def kamino_idl():
    """Load Kamino Lend v4 IDL (old format)."""
    idl_path = Path(__file__).parent.parent.parent / "kamino_lend_v4.json"
    with open(idl_path) as f:
        return json.load(f)


@pytest.fixture
def adrena_idl():
    """Load Adrena IDL (old format)."""
    idl_path = Path(__file__).parent.parent.parent / "adrena.json"
    with open(idl_path) as f:
        return json.load(f)


def test_new_format_idl_parsing(loopscale_idl):
    """Test parsing of new format IDL with precomputed discriminators."""
    # Parse the IDL
    idl = Idl.from_json(json.dumps(loopscale_idl))

    # Check accounts have discriminators
    assert len(idl.accounts) == 10
    for account in idl.accounts:
        # New format accounts should have discriminators
        if isinstance(account, dict):
            assert 'discriminator' in account
            assert len(account['discriminator']) == 8

    # Check types
    assert len(idl.types) == 69

    # Check for new fields like serialization and repr
    types_with_serialization = [
        t for t in idl.types
        if hasattr(t, 'serialization') and t.serialization
    ]
    assert len(types_with_serialization) > 0

    types_with_repr = [
        t for t in idl.types
        if hasattr(t, 'repr') and t.repr
    ]
    assert len(types_with_repr) > 0

    # Check address field
    assert idl.address == "1oopBoJG58DgkUVKkEzKgyG9dvRmpgeEm1AVjoHkF78"


def test_old_format_compatibility(kamino_idl):
    """Test that old format IDLs still work."""
    # Parse the IDL
    idl = Idl.from_json(json.dumps(kamino_idl))

    # Check basic structure
    assert idl.accounts is not None
    assert idl.types is not None
    assert idl.instructions is not None

    # Old format won't have address at top level
    assert idl.address is None or idl.address == ""


def test_program_creation_new_format(loopscale_idl):
    """Test creating Program with new format IDL."""
    provider = Provider.local()
    program_id = loopscale_idl['address']

    # Should be able to create program with dict IDL
    program = Program(loopscale_idl, program_id, provider)

    assert program.program_id
    assert program.idl
    assert program.provider == provider


def test_tuple_struct_support(loopscale_idl):
    """Test support for tuple structs (structs with unnamed fields)."""
    idl = Idl.from_json(json.dumps(loopscale_idl))

    # Find PodBool type which has tuple fields
    pod_bool = None
    for type_def in idl.types:
        if hasattr(type_def, 'name') and type_def.name == 'PodBool':
            pod_bool = type_def
            break

    assert pod_bool is not None
    # PodBool has fields: ['u8'] which is a tuple struct


def test_enum_variant_formats(loopscale_idl):
    """Test support for different enum variant formats."""
    idl = Idl.from_json(json.dumps(loopscale_idl))

    # Find TimelockUpdateParams which has mixed variant types
    timelock_enum = None
    for type_def in idl.types:
        if hasattr(type_def, 'name') and type_def.name == 'TimelockUpdateParams':
            timelock_enum = type_def
            break

    assert timelock_enum is not None
    # Should have enum type with variants that have different field formats


def test_defined_type_formats(loopscale_idl):
    """Test support for new defined type format."""
    idl = Idl.from_json(json.dumps(loopscale_idl))

    # New format uses {"name": "TypeName"} for defined types
    # Old format uses just "TypeName"
    # Both should work after our updates

    # Check that types with defined references work
    for type_def in idl.types:
        if hasattr(type_def, 'ty') and hasattr(type_def.ty, 'fields'):
            fields = type_def.ty.fields
            # Fields should be properly parsed regardless of format
            assert fields is not None


def test_discriminator_handling():
    """Test discriminator handling for both formats."""
    from anchorpy.coder.accounts import _account_discriminator

    # Test calculated discriminator (old format)
    calc_disc = _account_discriminator("MyAccount")
    assert len(calc_disc) == 8
    assert isinstance(calc_disc, bytes)

    # Test precomputed discriminator (new format)
    precomputed = [1, 2, 3, 4, 5, 6, 7, 8]
    disc_bytes = bytes(precomputed)
    assert len(disc_bytes) == 8


def test_backward_compatibility_fields():
    """Test field name compatibility (isMut/isSigner vs writable/signer)."""
    from anchorpy.coder.idl_compat import get_account_writable, get_account_signer

    # Old format
    old_account = type('Account', (), {'isMut': True, 'isSigner': False})()
    assert get_account_writable(old_account) == True
    assert get_account_signer(old_account) == False

    # New format
    new_account = type('Account', (), {'writable': True, 'signer': False})()
    assert get_account_writable(new_account) == True
    assert get_account_signer(new_account) == False

    # Dict format
    dict_account = {'writable': True, 'signer': True}
    assert get_account_writable(dict_account) == True
    assert get_account_signer(dict_account) == True


if __name__ == "__main__":
    pytest.main([__file__, "-v"])


"""IDL compatibility helpers for supporting both old and new Anchor IDL formats."""
from typing import Any, Dict, List, Optional, Union


def detect_idl_format(idl: Any) -> str:
    """Detect IDL format version.

    Args:
        idl: The IDL object.

    Returns:
        "old" for legacy format, "new" for spec 0.1.0+
    """
    # New format has 'spec' field in metadata
    if hasattr(idl, 'metadata') and idl.metadata and hasattr(idl.metadata, 'spec'):
        return "new"

    # New format has 'address' at top level
    if hasattr(idl, 'address') and idl.address:
        return "new"

    # Check for discriminator in first instruction (if any)
    if hasattr(idl, 'instructions') and idl.instructions:
        first_ix = idl.instructions[0]
        if hasattr(first_ix, 'discriminator') and first_ix.discriminator:
            return "new"

    # Check for discriminator in first account (if any)
    if hasattr(idl, 'accounts') and idl.accounts:
        first_acc = idl.accounts[0]
        if hasattr(first_acc, 'discriminator') and first_acc.discriminator:
            return "new"

    return "old"


def get_account_discriminator(account: Any) -> Optional[List[int]]:
    """Get precomputed discriminator from account if available.

    Args:
        account: The account object from IDL.

    Returns:
        List of 8 integers representing discriminator, or None if not present.
    """
    if hasattr(account, 'discriminator') and account.discriminator:
        return account.discriminator
    elif isinstance(account, dict) and 'discriminator' in account:
        return account['discriminator']
    return None


def get_instruction_discriminator(instruction: Any) -> Optional[List[int]]:
    """Get precomputed discriminator from instruction if available.

    Args:
        instruction: The instruction object from IDL.

    Returns:
        List of 8 integers representing discriminator, or None if not present.
    """
    if hasattr(instruction, 'discriminator') and instruction.discriminator:
        return instruction.discriminator
    return None


def get_event_discriminator(event: Any) -> Optional[List[int]]:
    """Get precomputed discriminator from event if available.

    Args:
        event: The event object from IDL.

    Returns:
        List of 8 integers representing discriminator, or None if not present.
    """
    if hasattr(event, 'discriminator') and event.discriminator:
        return event.discriminator
    return None


def get_account_writable(account: Any) -> bool:
    """Get writable flag with backward compatibility.

    Args:
        account: The account object from IDL instruction.

    Returns:
        True if account is writable/mutable, False otherwise.
    """
    # New format uses 'writable'
    if hasattr(account, 'writable'):
        return bool(account.writable)
    # Old format uses 'isMut'
    if hasattr(account, 'isMut'):
        return bool(account.isMut)
    # Default to False if neither field exists
    return False


def get_account_signer(account: Any) -> bool:
    """Get signer flag with backward compatibility.

    Args:
        account: The account object from IDL instruction.

    Returns:
        True if account is a signer, False otherwise.
    """
    # New format uses 'signer'
    if hasattr(account, 'signer'):
        return bool(account.signer)
    # Old format uses 'isSigner'
    if hasattr(account, 'isSigner'):
        return bool(account.isSigner)
    # Default to False if neither field exists
    return False


def get_account_address(account: Any) -> Optional[str]:
    """Get constant address from account if available.

    Args:
        account: The account object from IDL instruction.

    Returns:
        The address string if present (new format), None otherwise.
    """
    if hasattr(account, 'address') and account.address:
        return account.address
    return None


def get_defined_type_name(type_defined: Any) -> str:
    """Get the name of a defined type with backward compatibility.

    Args:
        type_defined: The defined type, either a string (old) or object with 'name' (new).

    Returns:
        The type name string.

    Raises:
        ValueError: If the type format is not recognized.
    """
    # Old format: defined is a string
    if isinstance(type_defined, str):
        return type_defined

    # New format: defined is an object with 'name' field
    if hasattr(type_defined, 'name'):
        return type_defined.name

    # Also handle dict format (for JSON-parsed IDLs)
    if isinstance(type_defined, dict) and 'name' in type_defined:
        return type_defined['name']

    raise ValueError(f"Unknown defined type format: {type_defined}")


def normalize_defined_type(type_obj: Any) -> Any:
    """Normalize defined type to handle both old and new formats.

    This updates the type object in place if needed.

    Args:
        type_obj: The type object that may contain a 'defined' field.

    Returns:
        The same type object (possibly modified).
    """
    if hasattr(type_obj, 'defined'):
        # If defined is already a string (old format), leave it as is
        if isinstance(type_obj.defined, str):
            return type_obj

        # If defined is an object with 'name' (new format), we may need to handle it
        # depending on how the anchorpy_core types are structured
        # For now, we'll leave it as is and handle in the specific code locations

    return type_obj


def get_account_type_definition(account: Any, types: Optional[List[Any]] = None) -> Any:
    """Get the type definition for an account.

    In the old format, accounts contain their full type definition.
    In the new format, accounts only have name + discriminator, and the
    type definition must be looked up in the types array.

    Args:
        account: The account object from the IDL accounts array.
        types: The types array from the IDL (needed for new format).

    Returns:
        The type definition for the account, or the account itself if it's old format.

    Raises:
        ValueError: If the type definition cannot be found.
    """
    # Handle both object and dict forms
    acc_name = None
    if hasattr(account, 'name'):
        acc_name = account.name
    elif isinstance(account, dict) and 'name' in account:
        acc_name = account['name']

    # Old format: account has a 'ty' field (it IS the type definition)
    if hasattr(account, 'ty'):
        return account
    elif isinstance(account, dict) and 'ty' in account:
        # Dict version of old format
        return account

    # Old format alternative: account has 'type' field
    if hasattr(account, 'type') and account.type:
        return account
    elif isinstance(account, dict) and 'type' in account:
        return account

    # Old format alternative: account has 'kind' field (it IS the type definition)
    if hasattr(account, 'kind'):
        return account
    elif isinstance(account, dict) and 'kind' in account:
        return account

    # New format: account only has name + discriminator, look up in types
    if types is None:
        raise ValueError(f"Types array required for new format account: {acc_name}")

    # Find the type definition by matching name
    for type_def in types:
        type_name = None
        if hasattr(type_def, 'name'):
            type_name = type_def.name
        elif isinstance(type_def, dict) and 'name' in type_def:
            type_name = type_def['name']

        if type_name == acc_name:
            return type_def

    raise ValueError(f"Type definition not found for account: {acc_name}")


from pathlib import Path

from anchorpy import Idl
from anchorpy.clientgen.instructions import gen_accounts
from anchorpy.clientgen.types import gen_struct
from genpy import Suite


def test_gen_accounts() -> None:
    path = Path("tests/idls/composite.json")
    raw = path.read_text()
    idl = Idl.from_json(raw)
    accs = gen_accounts(
        "CompositeUpdateAccounts", idl.instructions[1].accounts, gen_pdas=True
    )[0]
    suite = Suite(accs)
    assert str(suite) == (
        "    class CompositeUpdateAccounts(typing.TypedDict):"
        "\n        foo: FooNested"
        "\n        bar: BarNested"
        "\n    class FooNested(typing.TypedDict):"
        "\n        dummy_a: Pubkey"
        "\n    class BarNested(typing.TypedDict):"
        "\n        dummy_b: Pubkey"
        "\n    class FooNested(typing.TypedDict):"
        "\n        dummy_a: Pubkey"
    )


def test_empty_fields() -> None:
    path = Path("tests/idls/switchboard_v2.mainnet.06022022.json")
    raw = path.read_text()
    idl = Idl.from_json(raw)
    struct = gen_struct(idl, "AggregatorLockParams", [])
    assert str(struct) == (
        "import typing"
        "\nfrom dataclasses import dataclass"
        "\nfrom construct import Container, Construct"
        "\nfrom solders.pubkey import Pubkey"
        "\nfrom anchorpy.borsh_extension import BorshPubkey"
        "\nimport borsh_construct as borsh"
        "\nclass AggregatorLockParamsJSON(typing.TypedDict):"
        "\n    pass"
        "\n@dataclass"
        "\nclass AggregatorLockParams():"
        "\n    layout: typing.ClassVar = borsh.CStruct()"
        "\n    @classmethod"
        '\n    def from_decoded(cls, obj: Container) -> "AggregatorLockParams":'
        "\n        return cls()"
        "\n    def to_encodable(self) -> dict[str, typing.Any]:"
        "\n        return {}"
        "\n    def to_json(self) -> AggregatorLockParamsJSON:"
        "\n        return {}"
        "\n    @classmethod"
        '\n    def from_json(cls, obj: AggregatorLockParamsJSON) -> "AggregatorLockParams":'
        "\n        return cls()"
    )



"""This module deals (de)serializing program instructions."""
from typing import Any, Dict, Protocol, Tuple, TypeVar, cast

from anchorpy3_core.idl import Idl
from borsh_construct import CStruct
from construct import Adapter, Bytes, Construct, Container, Sequence, Switch
from pyheck import snake

from anchorpy.coder.common import _sighash
from anchorpy.coder.idl import _field_layout
from anchorpy.coder.idl_compat import get_instruction_discriminator
from anchorpy.idl import TypeDefs
from anchorpy.program.common import NamedInstruction


class _Sighash(Adapter):
    """Sighash as a Construct Adapter."""

    def __init__(self) -> None:
        """Initialize."""
        super().__init__(Bytes(8))  # type: ignore

    def _encode(self, obj: str, context, path) -> bytes:
        return _sighash(obj)

    def _decode(self, obj: bytes, context, path):
        raise ValueError("Sighash cannot be reversed")


class InstructionCoder(Adapter):
    """Encodes and decodes program instructions."""

    def __init__(self, idl: Idl) -> None:
        """Init.

        Args:
            idl: The parsed IDL object.
        """
        self.ix_layout = _parse_ix_layout(idl)
        sighasher = _Sighash()
        sighash_layouts: Dict[bytes, Construct] = {}
        sighashes: Dict[str, bytes] = {}
        sighash_to_name: Dict[bytes, str] = {}
        for ix in idl.instructions:
            ix_name = snake(ix.name)
            # Support both calculated and precomputed discriminators
            disc_list = get_instruction_discriminator(ix)
            if disc_list:
                # New format: use precomputed discriminator
                sh = bytes(disc_list)
            else:
                # Old format: calculate from name
                sh = sighasher.build(ix_name)
            sighashes[ix_name] = sh
            sighash_layouts[sh] = self.ix_layout[ix_name]
            sighash_to_name[sh] = ix_name
        self.sighash_layouts = sighash_layouts
        self.sighashes = sighashes
        self.sighash_to_name = sighash_to_name
        subcon = Sequence(
            "sighash" / Bytes(8),
            Switch(lambda this: this.sighash, sighash_layouts),
        )
        super().__init__(subcon)  # type: ignore

    def encode(self, ix_name: str, ix: Dict[str, Any]) -> bytes:
        """Encode a program instruction.

        Args:
            ix_name: The name of the instruction
            ix: The data to encode.

        Returns:
            The encoded instruction.
        """
        return self.build(NamedInstruction(name=ix_name, data=ix))

    def _decode(self, obj: Tuple[bytes, Any], context, path) -> NamedInstruction:
        return NamedInstruction(data=obj[1], name=self.sighash_to_name[obj[0]])

    def _encode(
        self, obj: NamedInstruction, context: Container, path
    ) -> Tuple[bytes, Any]:
        return (self.sighashes[obj.name], obj.data)


_SA = TypeVar("_SA", bound="_SupportsAdd")


class _SupportsAdd(Protocol):
    """Any type T where +(:T, :T) -> T."""

    def __add__(self: _SA, other: _SA) -> _SA:
        ...


def _parse_ix_layout(idl: Idl) -> Dict[str, Construct]:
    ix_layout: Dict[str, Construct] = {}
    for ix in idl.instructions:
        typedefs = cast(_SupportsAdd, idl.accounts) + cast(_SupportsAdd, idl.types)
        field_layouts = [
            _field_layout(arg, cast(TypeDefs, typedefs)) for arg in ix.args
        ]
        ix_name = snake(ix.name)
        ix_layout[ix_name] = ix_name / CStruct(*field_layouts)
    return ix_layout



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class BarStructJSON(typing.TypedDict):
    some_field: bool
    other_field: int


@dataclass
class BarStruct:
    layout: typing.ClassVar = borsh.CStruct(
        "some_field" / borsh.Bool, "other_field" / borsh.U8
    )
    some_field: bool
    other_field: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "BarStruct":
        return cls(some_field=obj.some_field, other_field=obj.other_field)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"some_field": self.some_field, "other_field": self.other_field}

    def to_json(self) -> BarStructJSON:
        return {"some_field": self.some_field, "other_field": self.other_field}

    @classmethod
    def from_json(cls, obj: BarStructJSON) -> "BarStruct":
        return cls(some_field=obj["some_field"], other_field=obj["other_field"])



"""Extensions to the Borsh spec for Solana-specific types."""
from dataclasses import asdict
from keyword import kwlist
from typing import Any, Dict, Type, TypeVar, cast

from borsh_construct import U8, CStruct
from construct import (
    Adapter,
    Bytes,
    Construct,
    Container,
    IfThenElse,
    Padding,
    Renamed,
    Switch,
)
from solders import pubkey


class BorshPubkeyAdapter(Adapter):
    def __init__(self) -> None:
        super().__init__(Bytes(32))  # type: ignore

    def _decode(self, obj: bytes, context, path) -> pubkey.Pubkey:
        return pubkey.Pubkey(obj)

    def _encode(self, obj: pubkey.Pubkey, context, path) -> bytes:
        return bytes(obj)


class EnumForCodegen(Adapter):
    _index_key = "index"
    _value_key = "value"

    def __init__(self, *variants: "Renamed[CStruct, CStruct]") -> None:
        """Init enum."""
        switch_cases: dict[int, "Renamed[CStruct, CStruct]"] = {}
        variant_name_to_index: dict[str, int] = {}
        index_to_variant_name: dict[int, str] = {}
        for idx, parser in enumerate(variants):
            switch_cases[idx] = parser
            name = cast(str, parser.name)
            variant_name_to_index[name] = idx
            index_to_variant_name[idx] = name
        enum_struct = CStruct(
            self._index_key / U8,
            self._value_key
            / Switch(lambda this: this.index, cast(dict[int, Construct], switch_cases)),
        )
        super().__init__(enum_struct)  # type: ignore
        self.variant_name_to_index = variant_name_to_index
        self.index_to_variant_name = index_to_variant_name

    def _decode(self, obj: CStruct, context, path) -> dict[str, Any]:
        index = obj.index
        variant_name = self.index_to_variant_name[index]
        return {variant_name: obj.value}

    def _encode(self, obj: dict[str, Any], context, path) -> dict[str, Any]:
        variant_name = list(obj.keys())[0]
        index = self.variant_name_to_index[variant_name]
        return {self._index_key: index, self._value_key: obj[variant_name]}


class COption(Adapter):
    _discriminator_key = "discriminator"
    _value_key = "value"

    def __init__(self, subcon: Construct) -> None:
        option_struct = CStruct(
            self._discriminator_key / U8,
            self._value_key
            / IfThenElse(
                lambda this: this[self._discriminator_key] == 0,
                Padding(subcon.sizeof()),
                subcon,
            ),
        )
        super().__init__(option_struct)  # type: ignore

    def _decode(self, obj, context, path) -> Any:
        discriminator = obj[self._discriminator_key]
        return None if discriminator == 0 else obj[self._value_key]

    def _encode(self, obj, context, path) -> dict:
        discriminator = 0 if obj is None else 1
        return {self._discriminator_key: discriminator, self._value_key: obj}


T = TypeVar("T")


class _DataclassStruct(Adapter):
    """Converts dataclasses to/from `borsh_construct.CStruct`."""

    def __init__(self, cstruct: CStruct, datacls: Type[T]) -> None:
        """Init.

        Args:
            cstruct: The underlying `CStruct`.
            datacls: The dataclass type.
        """
        super().__init__(cstruct)  # type: ignore
        self.datacls = datacls

    def _decode(self, obj: Container, context, path) -> T:  # type: ignore
        kwargs = {}
        for key, value in obj.items():
            if key[0] != "_":
                key_to_use = f"{key}_" if key in kwlist else key
                kwargs[key_to_use] = value
        return self.datacls(**kwargs)  # type: ignore

    def _encode(self, obj: T, context, path) -> Dict[str, Any]:
        if isinstance(obj, dict):
            return obj
        return asdict(obj)


BorshPubkey = BorshPubkeyAdapter()
"""Adapter for (de)serializing a public key."""



from __future__ import annotations
from . import bar_struct
import typing
from dataclasses import dataclass
from construct import Construct
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh

UnnamedJSONValue = tuple[bool, int, bar_struct.BarStructJSON]
UnnamedSingleJSONValue = tuple[bar_struct.BarStructJSON]


class NamedJSONValue(typing.TypedDict):
    bool_field: bool
    u8_field: int
    nested: bar_struct.BarStructJSON


StructJSONValue = tuple[bar_struct.BarStructJSON]
OptionStructJSONValue = tuple[typing.Optional[bar_struct.BarStructJSON]]
VecStructJSONValue = tuple[list[bar_struct.BarStructJSON]]
UnnamedValue = tuple[bool, int, bar_struct.BarStruct]
UnnamedSingleValue = tuple[bar_struct.BarStruct]


class NamedValue(typing.TypedDict):
    bool_field: bool
    u8_field: int
    nested: bar_struct.BarStruct


StructValue = tuple[bar_struct.BarStruct]
OptionStructValue = tuple[typing.Optional[bar_struct.BarStruct]]
VecStructValue = tuple[list[bar_struct.BarStruct]]


class UnnamedJSON(typing.TypedDict):
    value: UnnamedJSONValue
    kind: typing.Literal["Unnamed"]


class UnnamedSingleJSON(typing.TypedDict):
    value: UnnamedSingleJSONValue
    kind: typing.Literal["UnnamedSingle"]


class NamedJSON(typing.TypedDict):
    value: NamedJSONValue
    kind: typing.Literal["Named"]


class StructJSON(typing.TypedDict):
    value: StructJSONValue
    kind: typing.Literal["Struct"]


class OptionStructJSON(typing.TypedDict):
    value: OptionStructJSONValue
    kind: typing.Literal["OptionStruct"]


class VecStructJSON(typing.TypedDict):
    value: VecStructJSONValue
    kind: typing.Literal["VecStruct"]


class NoFieldsJSON(typing.TypedDict):
    kind: typing.Literal["NoFields"]


@dataclass
class Unnamed:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Unnamed"
    value: UnnamedValue

    def to_json(self) -> UnnamedJSON:
        return UnnamedJSON(
            kind="Unnamed",
            value=(
                self.value[0],
                self.value[1],
                self.value[2].to_json(),
            ),
        )

    def to_encodable(self) -> dict:
        return {
            "Unnamed": {
                "item_0": self.value[0],
                "item_1": self.value[1],
                "item_2": self.value[2].to_encodable(),
            },
        }


@dataclass
class UnnamedSingle:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "UnnamedSingle"
    value: UnnamedSingleValue

    def to_json(self) -> UnnamedSingleJSON:
        return UnnamedSingleJSON(
            kind="UnnamedSingle",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "UnnamedSingle": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Named:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Named"
    value: NamedValue

    def to_json(self) -> NamedJSON:
        return NamedJSON(
            kind="Named",
            value={
                "bool_field": self.value["bool_field"],
                "u8_field": self.value["u8_field"],
                "nested": self.value["nested"].to_json(),
            },
        )

    def to_encodable(self) -> dict:
        return {
            "Named": {
                "bool_field": self.value["bool_field"],
                "u8_field": self.value["u8_field"],
                "nested": self.value["nested"].to_encodable(),
            },
        }


@dataclass
class Struct:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Struct"
    value: StructValue

    def to_json(self) -> StructJSON:
        return StructJSON(
            kind="Struct",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Struct": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class OptionStruct:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "OptionStruct"
    value: OptionStructValue

    def to_json(self) -> OptionStructJSON:
        return OptionStructJSON(
            kind="OptionStruct",
            value=((None if self.value[0] is None else self.value[0].to_json()),),
        )

    def to_encodable(self) -> dict:
        return {
            "OptionStruct": {
                "item_0": (
                    None if self.value[0] is None else self.value[0].to_encodable()
                ),
            },
        }


@dataclass
class VecStruct:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "VecStruct"
    value: VecStructValue

    def to_json(self) -> VecStructJSON:
        return VecStructJSON(
            kind="VecStruct",
            value=(list(map(lambda item: item.to_json(), self.value[0])),),
        )

    def to_encodable(self) -> dict:
        return {
            "VecStruct": {
                "item_0": list(map(lambda item: item.to_encodable(), self.value[0])),
            },
        }


@dataclass
class NoFields:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "NoFields"

    @classmethod
    def to_json(cls) -> NoFieldsJSON:
        return NoFieldsJSON(
            kind="NoFields",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "NoFields": {},
        }


FooEnumKind = typing.Union[
    Unnamed, UnnamedSingle, Named, Struct, OptionStruct, VecStruct, NoFields
]
FooEnumJSON = typing.Union[
    UnnamedJSON,
    UnnamedSingleJSON,
    NamedJSON,
    StructJSON,
    OptionStructJSON,
    VecStructJSON,
    NoFieldsJSON,
]


def from_decoded(obj: dict) -> FooEnumKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Unnamed" in obj:
        val = obj["Unnamed"]
        return Unnamed(
            (
                val["item_0"],
                val["item_1"],
                bar_struct.BarStruct.from_decoded(val["item_2"]),
            )
        )
    if "UnnamedSingle" in obj:
        val = obj["UnnamedSingle"]
        return UnnamedSingle((bar_struct.BarStruct.from_decoded(val["item_0"]),))
    if "Named" in obj:
        val = obj["Named"]
        return Named(
            NamedValue(
                bool_field=val["bool_field"],
                u8_field=val["u8_field"],
                nested=bar_struct.BarStruct.from_decoded(val["nested"]),
            )
        )
    if "Struct" in obj:
        val = obj["Struct"]
        return Struct((bar_struct.BarStruct.from_decoded(val["item_0"]),))
    if "OptionStruct" in obj:
        val = obj["OptionStruct"]
        return OptionStruct(
            (
                (
                    None
                    if val["item_0"] is None
                    else bar_struct.BarStruct.from_decoded(val["item_0"])
                ),
            )
        )
    if "VecStruct" in obj:
        val = obj["VecStruct"]
        return VecStruct(
            (
                list(
                    map(
                        lambda item: bar_struct.BarStruct.from_decoded(item),
                        val["item_0"],
                    )
                ),
            )
        )
    if "NoFields" in obj:
        return NoFields()
    raise ValueError("Invalid enum object")


def from_json(obj: FooEnumJSON) -> FooEnumKind:
    if obj["kind"] == "Unnamed":
        unnamed_json_value = typing.cast(UnnamedJSONValue, obj["value"])
        return Unnamed(
            (
                unnamed_json_value[0],
                unnamed_json_value[1],
                bar_struct.BarStruct.from_json(unnamed_json_value[2]),
            )
        )
    if obj["kind"] == "UnnamedSingle":
        unnamed_single_json_value = typing.cast(UnnamedSingleJSONValue, obj["value"])
        return UnnamedSingle(
            (bar_struct.BarStruct.from_json(unnamed_single_json_value[0]),)
        )
    if obj["kind"] == "Named":
        named_json_value = typing.cast(NamedJSONValue, obj["value"])
        return Named(
            NamedValue(
                bool_field=named_json_value["bool_field"],
                u8_field=named_json_value["u8_field"],
                nested=bar_struct.BarStruct.from_json(named_json_value["nested"]),
            )
        )
    if obj["kind"] == "Struct":
        struct_json_value = typing.cast(StructJSONValue, obj["value"])
        return Struct((bar_struct.BarStruct.from_json(struct_json_value[0]),))
    if obj["kind"] == "OptionStruct":
        option_struct_json_value = typing.cast(OptionStructJSONValue, obj["value"])
        return OptionStruct(
            (
                (
                    None
                    if option_struct_json_value[0] is None
                    else bar_struct.BarStruct.from_json(option_struct_json_value[0])
                ),
            )
        )
    if obj["kind"] == "VecStruct":
        vec_struct_json_value = typing.cast(VecStructJSONValue, obj["value"])
        return VecStruct(
            (
                list(
                    map(
                        lambda item: bar_struct.BarStruct.from_json(item),
                        vec_struct_json_value[0],
                    )
                ),
            )
        )
    if obj["kind"] == "NoFields":
        return NoFields()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Unnamed"
    / borsh.CStruct(
        "item_0" / borsh.Bool,
        "item_1" / borsh.U8,
        "item_2" / bar_struct.BarStruct.layout,
    ),
    "UnnamedSingle" / borsh.CStruct("item_0" / bar_struct.BarStruct.layout),
    "Named"
    / borsh.CStruct(
        "bool_field" / borsh.Bool,
        "u8_field" / borsh.U8,
        "nested" / bar_struct.BarStruct.layout,
    ),
    "Struct" / borsh.CStruct("item_0" / bar_struct.BarStruct.layout),
    "OptionStruct"
    / borsh.CStruct("item_0" / borsh.Option(bar_struct.BarStruct.layout)),
    "VecStruct"
    / borsh.CStruct(
        "item_0" / borsh.Vec(typing.cast(Construct, bar_struct.BarStruct.layout))
    ),
    "NoFields" / borsh.CStruct(),
)



from pathlib import Path

from anchorpy import AccountsCoder, Idl
from pytest import mark


@mark.unit
def test_accounts_coder() -> None:
    """Test accounts coder."""
    raw = Path("tests/idls/basic_1.json").read_text()
    idl = Idl.from_json(raw)
    raw_acc_data = b"\xf6\x1c\x06W\xfb-2*\xd2\x04\x00\x00\x00\x00\x00\x00"
    acc_coder = AccountsCoder(idl)
    decoded = acc_coder.parse(raw_acc_data)
    encoded = acc_coder.build(decoded)
    assert encoded == raw_acc_data



import typing
from dataclasses import dataclass
from construct import Construct
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from .. import types


class StateJSON(typing.TypedDict):
    bool_field: bool
    u8_field: int
    i8_field: int
    u16_field: int
    i16_field: int
    u32_field: int
    i32_field: int
    f32_field: float
    u64_field: int
    i64_field: int
    f64_field: float
    u128_field: int
    i128_field: int
    bytes_field: list[int]
    string_field: str
    pubkey_field: str
    vec_field: list[int]
    vec_struct_field: list[types.foo_struct.FooStructJSON]
    option_field: typing.Optional[bool]
    option_struct_field: typing.Optional[types.foo_struct.FooStructJSON]
    struct_field: types.foo_struct.FooStructJSON
    array_field: list[bool]
    enum_field1: types.foo_enum.FooEnumJSON
    enum_field2: types.foo_enum.FooEnumJSON
    enum_field3: types.foo_enum.FooEnumJSON
    enum_field4: types.foo_enum.FooEnumJSON


@dataclass
class State:
    discriminator: typing.ClassVar = b"\xd8\x92k^hK\xb6\xb1"
    layout: typing.ClassVar = borsh.CStruct(
        "bool_field" / borsh.Bool,
        "u8_field" / borsh.U8,
        "i8_field" / borsh.I8,
        "u16_field" / borsh.U16,
        "i16_field" / borsh.I16,
        "u32_field" / borsh.U32,
        "i32_field" / borsh.I32,
        "f32_field" / borsh.F32,
        "u64_field" / borsh.U64,
        "i64_field" / borsh.I64,
        "f64_field" / borsh.F64,
        "u128_field" / borsh.U128,
        "i128_field" / borsh.I128,
        "bytes_field" / borsh.Bytes,
        "string_field" / borsh.String,
        "pubkey_field" / BorshPubkey,
        "vec_field" / borsh.Vec(typing.cast(Construct, borsh.U64)),
        "vec_struct_field"
        / borsh.Vec(typing.cast(Construct, types.foo_struct.FooStruct.layout)),
        "option_field" / borsh.Option(borsh.Bool),
        "option_struct_field" / borsh.Option(types.foo_struct.FooStruct.layout),
        "struct_field" / types.foo_struct.FooStruct.layout,
        "array_field" / borsh.Bool[3],
        "enum_field1" / types.foo_enum.layout,
        "enum_field2" / types.foo_enum.layout,
        "enum_field3" / types.foo_enum.layout,
        "enum_field4" / types.foo_enum.layout,
    )
    bool_field: bool
    u8_field: int
    i8_field: int
    u16_field: int
    i16_field: int
    u32_field: int
    i32_field: int
    f32_field: float
    u64_field: int
    i64_field: int
    f64_field: float
    u128_field: int
    i128_field: int
    bytes_field: bytes
    string_field: str
    pubkey_field: Pubkey
    vec_field: list[int]
    vec_struct_field: list[types.foo_struct.FooStruct]
    option_field: typing.Optional[bool]
    option_struct_field: typing.Optional[types.foo_struct.FooStruct]
    struct_field: types.foo_struct.FooStruct
    array_field: list[bool]
    enum_field1: types.foo_enum.FooEnumKind
    enum_field2: types.foo_enum.FooEnumKind
    enum_field3: types.foo_enum.FooEnumKind
    enum_field4: types.foo_enum.FooEnumKind

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["State"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["State"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["State"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "State":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = State.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            bool_field=dec.bool_field,
            u8_field=dec.u8_field,
            i8_field=dec.i8_field,
            u16_field=dec.u16_field,
            i16_field=dec.i16_field,
            u32_field=dec.u32_field,
            i32_field=dec.i32_field,
            f32_field=dec.f32_field,
            u64_field=dec.u64_field,
            i64_field=dec.i64_field,
            f64_field=dec.f64_field,
            u128_field=dec.u128_field,
            i128_field=dec.i128_field,
            bytes_field=dec.bytes_field,
            string_field=dec.string_field,
            pubkey_field=dec.pubkey_field,
            vec_field=dec.vec_field,
            vec_struct_field=list(
                map(
                    lambda item: types.foo_struct.FooStruct.from_decoded(item),
                    dec.vec_struct_field,
                )
            ),
            option_field=dec.option_field,
            option_struct_field=(
                None
                if dec.option_struct_field is None
                else types.foo_struct.FooStruct.from_decoded(dec.option_struct_field)
            ),
            struct_field=types.foo_struct.FooStruct.from_decoded(dec.struct_field),
            array_field=dec.array_field,
            enum_field1=types.foo_enum.from_decoded(dec.enum_field1),
            enum_field2=types.foo_enum.from_decoded(dec.enum_field2),
            enum_field3=types.foo_enum.from_decoded(dec.enum_field3),
            enum_field4=types.foo_enum.from_decoded(dec.enum_field4),
        )

    def to_json(self) -> StateJSON:
        return {
            "bool_field": self.bool_field,
            "u8_field": self.u8_field,
            "i8_field": self.i8_field,
            "u16_field": self.u16_field,
            "i16_field": self.i16_field,
            "u32_field": self.u32_field,
            "i32_field": self.i32_field,
            "f32_field": self.f32_field,
            "u64_field": self.u64_field,
            "i64_field": self.i64_field,
            "f64_field": self.f64_field,
            "u128_field": self.u128_field,
            "i128_field": self.i128_field,
            "bytes_field": list(self.bytes_field),
            "string_field": self.string_field,
            "pubkey_field": str(self.pubkey_field),
            "vec_field": self.vec_field,
            "vec_struct_field": list(
                map(lambda item: item.to_json(), self.vec_struct_field)
            ),
            "option_field": self.option_field,
            "option_struct_field": (
                None
                if self.option_struct_field is None
                else self.option_struct_field.to_json()
            ),
            "struct_field": self.struct_field.to_json(),
            "array_field": self.array_field,
            "enum_field1": self.enum_field1.to_json(),
            "enum_field2": self.enum_field2.to_json(),
            "enum_field3": self.enum_field3.to_json(),
            "enum_field4": self.enum_field4.to_json(),
        }

    @classmethod
    def from_json(cls, obj: StateJSON) -> "State":
        return cls(
            bool_field=obj["bool_field"],
            u8_field=obj["u8_field"],
            i8_field=obj["i8_field"],
            u16_field=obj["u16_field"],
            i16_field=obj["i16_field"],
            u32_field=obj["u32_field"],
            i32_field=obj["i32_field"],
            f32_field=obj["f32_field"],
            u64_field=obj["u64_field"],
            i64_field=obj["i64_field"],
            f64_field=obj["f64_field"],
            u128_field=obj["u128_field"],
            i128_field=obj["i128_field"],
            bytes_field=bytes(obj["bytes_field"]),
            string_field=obj["string_field"],
            pubkey_field=Pubkey.from_string(obj["pubkey_field"]),
            vec_field=obj["vec_field"],
            vec_struct_field=list(
                map(
                    lambda item: types.foo_struct.FooStruct.from_json(item),
                    obj["vec_struct_field"],
                )
            ),
            option_field=obj["option_field"],
            option_struct_field=(
                None
                if obj["option_struct_field"] is None
                else types.foo_struct.FooStruct.from_json(obj["option_struct_field"])
            ),
            struct_field=types.foo_struct.FooStruct.from_json(obj["struct_field"]),
            array_field=obj["array_field"],
            enum_field1=types.foo_enum.from_json(obj["enum_field1"]),
            enum_field2=types.foo_enum.from_json(obj["enum_field2"]),
            enum_field3=types.foo_enum.from_json(obj["enum_field3"]),
            enum_field4=types.foo_enum.from_json(obj["enum_field4"]),
        )



"""This module contains the Provider class and associated utilities."""
from __future__ import annotations

import json
from os import environ, getenv
from pathlib import Path
from types import MappingProxyType
from typing import List, Optional, Sequence, Union

from solana.rpc import types
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Confirmed, Finalized, Processed
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.rpc.responses import SimulateTransactionResp
from solders.signature import Signature
from solders.transaction import Transaction, VersionedTransaction

DEFAULT_OPTIONS = types.TxOpts(skip_confirmation=False, preflight_commitment=Processed)
COMMITMENT_RANKS = MappingProxyType({Processed: 0, Confirmed: 1, Finalized: 2})


class Provider:
    """The network and wallet context used to send transactions paid for and signed by the provider."""  # noqa: E501

    def __init__(
        self,
        connection: AsyncClient,
        wallet: Wallet,
        opts: types.TxOpts = DEFAULT_OPTIONS,
    ) -> None:
        """Initialize the Provider.

        Args:
            connection: The cluster connection where the program is deployed.
            wallet: The wallet used to pay for and sign all transactions.
            opts: Transaction confirmation options to use by default.
        """
        self.connection = connection
        self.wallet = wallet
        self.opts = opts

    @classmethod
    def local(
        cls, url: Optional[str] = None, opts: types.TxOpts = DEFAULT_OPTIONS
    ) -> Provider:
        """Create a `Provider` with a wallet read from the local filesystem.

        Args:
            url: The network cluster url.
            opts: The default transaction confirmation options.
        """
        connection = AsyncClient(url, opts.preflight_commitment)
        wallet = Wallet.local()
        return cls(connection, wallet, opts)

    @classmethod
    def readonly(
        cls, url: Optional[str] = None, opts: types.TxOpts = DEFAULT_OPTIONS
    ) -> Provider:
        """Create a `Provider` that can only fetch data, not send transactions.

        Args:
            url: The network cluster url.
            opts: The default transaction confirmation options.
        """
        connection = AsyncClient(url, opts.preflight_commitment)
        wallet = Wallet.dummy()
        return cls(connection, wallet, opts)

    @classmethod
    def env(cls) -> Provider:
        """Create a `Provider` using the `ANCHOR_PROVIDER_URL` environment variable."""
        url = environ["ANCHOR_PROVIDER_URL"]
        options = DEFAULT_OPTIONS
        connection = AsyncClient(url, options.preflight_commitment)
        wallet = Wallet.local()
        return cls(connection, wallet, options)

    async def simulate(
        self,
        tx: Union[Transaction, VersionedTransaction],
        opts: Optional[types.TxOpts] = None,
    ) -> SimulateTransactionResp:
        """Simulate the given transaction, returning emitted logs from execution.

        Args:
            tx: The transaction to send.
            signers: The set of signers in addition to the provider wallet that will
                sign the transaction.
            opts: Transaction confirmation options.

        Returns:
            The transaction simulation result.
        """
        if opts is None:
            opts = self.opts
        return await self.connection.simulate_transaction(
            tx, sig_verify=True, commitment=opts.preflight_commitment
        )

    async def send(
        self,
        tx: Union[Transaction, VersionedTransaction],
        opts: Optional[types.TxOpts] = None,
    ) -> Signature:
        """Send the given transaction, paid for and signed by the provider's wallet.

        Args:
            tx: The transaction to send.
            signers: The set of signers in addition to the provider wallet that will
                sign the transaction.
            opts: Transaction confirmation options.

        Returns:
            The transaction signature from the RPC server.
        """
        if opts is None:
            opts = self.opts
        raw = bytes(tx)
        resp = await self.connection.send_raw_transaction(raw, opts=opts)
        return resp.value

    async def send_all(
        self,
        txs: Sequence[Union[Transaction, VersionedTransaction]],
        opts: Optional[types.TxOpts] = None,
    ) -> list[Signature]:
        """Similar to `send`, but for an array of transactions and signers.

        Args:
            txs: a list of transaction objects.
            opts: Transaction confirmation options.

        Returns:
            The transaction signatures from the RPC server.
        """
        if opts is None:
            opts = self.opts
        sigs = []
        for tx in txs:
            raw = bytes(tx)
            resp = await self.connection.send_raw_transaction(raw, opts=opts)
            sigs.append(resp.value)
        return sigs

    async def __aenter__(self) -> Provider:
        """Use as a context manager."""
        await self.connection.__aenter__()
        return self

    async def __aexit__(self, _exc_type, _exc, _tb):
        """Exit the context manager."""
        await self.close()

    async def close(self) -> None:
        """Use this when you are done with the connection."""
        await self.connection.close()


class Wallet:
    """Python wallet object."""

    def __init__(self, payer: Keypair):
        """Initialize the wallet.

        Args:
            payer: the Keypair used to sign transactions.
        """
        self.payer = payer

    @property
    def public_key(self) -> Pubkey:
        """Get the public key of the wallet."""
        return self.payer.pubkey()

    def sign_transaction(self, tx: Transaction) -> Transaction:
        """Sign a transaction using the wallet's keypair.

        Args:
            tx: The transaction to sign.

        Returns:
            The signed transaction.
        """
        tx.sign([self.payer], tx.message.recent_blockhash)
        return tx

    def sign_all_transactions(self, txs: list[Transaction]) -> list[Transaction]:
        """Sign a list of transactions using the wallet's keypair.

        Args:
            txs: The transactions to sign.

        Returns:
            The signed transactions.
        """
        for tx in txs:
            tx.partial_sign([self.payer], tx.message.recent_blockhash)
        return txs

    @classmethod
    def local(cls) -> Wallet:
        """Create a wallet instance from the filesystem.

        Uses the path at the ANCHOR_WALLET env var if set,
        otherwise uses ~/.config/solana/id.json.
        """
        path = Path(getenv("ANCHOR_WALLET", Path.home() / ".config/solana/id.json"))
        with path.open() as f:
            keypair: List[int] = json.load(f)
        return cls(Keypair.from_bytes(keypair))

    @classmethod
    def dummy(cls) -> Wallet:
        """Create a dummy wallet instance that won't be used to sign transactions."""
        keypair = Keypair()
        return cls(keypair)



"""Test that the CLI commands work."""

from pathlib import Path

from anchorpy import localnet_fixture
from anchorpy.cli import app
from solana.rpc.api import Client
from solana.rpc.commitment import Processed
from solders.signature import Signature
from typer.testing import CliRunner

PATH = Path("anchor/examples/tutorial/basic-0")

localnet = localnet_fixture(PATH)

runner = CliRunner()


def test_shell(localnet, monkeypatch) -> None:
    monkeypatch.chdir("anchor/examples/tutorial/basic-0")
    cli_input = "await workspace['basic_0'].rpc['initialize']()\nexit()"
    result = runner.invoke(app, ["shell"], input=cli_input)
    assert result.exit_code == 0
    assert "Hint: type `workspace`" in result.stdout
    tx_sig = Signature.from_string(
        result.stdout.split("Out[1]: \n")[1].split("\n    ")[1].split(",")[0]
    )
    client = Client()
    client.confirm_transaction(tx_sig, commitment=Processed)



"""Test real-world IDL files for compatibility."""
import json
from pathlib import Path
from typing import Optional

import pytest
from solders.pubkey import Pubkey

from anchorpy.coder.accounts import AccountsCoder
from anchorpy.coder.coder import Coder
from anchorpy.coder.instruction import InstructionCoder
from anchorpy.coder.idl_compat import detect_idl_format
from anchorpy import Program, Provider


def load_idl_if_exists(filename: str) -> Optional[dict]:
    """Load IDL file if it exists."""
    idl_path = Path(__file__).parent.parent.parent / filename
    if idl_path.exists():
        with open(idl_path) as f:
            return json.load(f)
    return None


class TestKaminoLendV4:
    """Test Kamino Lend V4 IDL (old format)."""

    def test_kamino_idl_loads(self):
        """Test that Kamino IDL can be loaded."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            assert idl_data is not None
            assert "version" in idl_data
            assert idl_data["name"] == "kamino_lending"

    def test_kamino_is_old_format(self):
        """Test that Kamino is detected as old format."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            # Check structure to determine format
            assert "version" in idl_data
            assert "name" in idl_data
            # Old format has version and name at top level
            assert "spec" not in idl_data.get("metadata", {})

    @pytest.mark.asyncio
    async def test_kamino_program_creation(self):
        """Test creating Program with Kamino IDL."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            provider = Provider.readonly()
            program_id = Pubkey.new_unique()

            # Create program with Kamino IDL
            program = Program(idl_data, program_id, provider)

            assert program is not None
            assert program.program_id == program_id

    @pytest.mark.asyncio
    async def test_kamino_instruction_encoding(self):
        """Test instruction encoding with Kamino IDL."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            from anchorpy3_core.idl import Idl

            idl = Idl.from_json(json.dumps(idl_data))
            coder = InstructionCoder(idl)

            # Verify we can encode an instruction
            if idl.instructions:
                first_ix = idl.instructions[0]
                from pyheck import snake
                ix_name = snake(first_ix.name)

                # Check that sighash was created
                assert ix_name in coder.sighashes
                assert len(coder.sighashes[ix_name]) == 8

    def test_kamino_account_discriminators(self):
        """Test account discriminators with Kamino IDL."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            from anchorpy3_core.idl import Idl

            idl = Idl.from_json(json.dumps(idl_data))
            coder = AccountsCoder(idl)

            # Verify discriminators were calculated
            if idl.accounts:
                for acc in idl.accounts:
                    assert acc.name in coder.acc_name_to_discriminator
                    disc = coder.acc_name_to_discriminator[acc.name]
                    assert isinstance(disc, bytes)
                    assert len(disc) == 8


class TestAdrena:
    """Test Adrena IDL (old format)."""

    def test_adrena_idl_loads(self):
        """Test that Adrena IDL can be loaded."""
        idl_data = load_idl_if_exists("adrena.json")
        if idl_data:
            assert idl_data is not None
            assert "version" in idl_data
            assert idl_data["name"] == "adrena"

    def test_adrena_is_old_format(self):
        """Test that Adrena is detected as old format."""
        idl_data = load_idl_if_exists("adrena.json")
        if idl_data:
            # Old format characteristics
            assert "version" in idl_data
            assert "name" in idl_data
            # Should not have new format fields
            assert "address" not in idl_data or "spec" not in idl_data.get("metadata", {})

    @pytest.mark.asyncio
    async def test_adrena_program_creation(self):
        """Test creating Program with Adrena IDL."""
        idl_data = load_idl_if_exists("adrena.json")
        if idl_data:
            provider = Provider.readonly()
            program_id = Pubkey.new_unique()

            # Create program with Adrena IDL
            program = Program(idl_data, program_id, provider)

            assert program is not None
            assert program.program_id == program_id

    def test_adrena_defined_types(self):
        """Test that Adrena's defined types are handled correctly."""
        idl_data = load_idl_if_exists("adrena.json")
        if idl_data:
            from anchorpy3_core.idl import Idl

            idl = Idl.from_json(json.dumps(idl_data))

            # Check that types exist and can be processed
            if idl.types:
                assert len(idl.types) > 0
                # Verify all types have names
                for t in idl.types:
                    assert hasattr(t, "name")
                    assert t.name is not None


class TestLoopscaleV1:
    """Test Loopscale V1 IDL (new format)."""

    def test_loopscale_idl_loads(self):
        """Test that Loopscale IDL can be loaded."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            assert idl_data is not None
            assert "address" in idl_data
            assert "metadata" in idl_data
            assert idl_data["metadata"]["name"] == "loopscale"

    def test_loopscale_is_new_format(self):
        """Test that Loopscale is detected as new format."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            # New format characteristics
            assert "address" in idl_data
            assert "metadata" in idl_data
            assert "spec" in idl_data["metadata"]
            assert idl_data["metadata"]["spec"] == "0.1.0"

    def test_loopscale_has_discriminators(self):
        """Test that Loopscale has precomputed discriminators."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            # Check accounts have discriminators
            if "accounts" in idl_data and idl_data["accounts"]:
                for acc in idl_data["accounts"]:
                    assert "discriminator" in acc
                    assert isinstance(acc["discriminator"], list)
                    assert len(acc["discriminator"]) == 8

    @pytest.mark.asyncio
    async def test_loopscale_program_creation(self):
        """Test creating Program with Loopscale IDL."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            provider = Provider.readonly()

            # Use the address from the IDL if present
            if "address" in idl_data:
                program_id = Pubkey.from_string(idl_data["address"])
            else:
                program_id = Pubkey.new_unique()

            # Create program with Loopscale IDL
            program = Program(idl_data, program_id, provider)

            assert program is not None
            assert program.program_id == program_id

    def test_loopscale_account_discriminators(self):
        """Test that Loopscale uses precomputed discriminators."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            from anchorpy3_core.idl import Idl

            # For new format, we need to check if anchorpy_core supports it
            # If not, we'll test the raw data
            if "accounts" in idl_data and idl_data["accounts"]:
                for acc in idl_data["accounts"]:
                    # Verify discriminator exists and is correct format
                    assert "discriminator" in acc
                    disc = acc["discriminator"]
                    assert isinstance(disc, list)
                    assert len(disc) == 8
                    # All values should be valid bytes (0-255)
                    for val in disc:
                        assert isinstance(val, int)
                        assert 0 <= val <= 255


class TestClientGeneration:
    """Test client generation with both old and new format IDLs."""

    @pytest.mark.asyncio
    async def test_client_gen_with_old_format(self):
        """Test client generation with old format IDL."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            # This would test the client generation
            # In a real test, we'd generate the client and verify the output
            assert True  # Placeholder for actual client gen test

    @pytest.mark.asyncio
    async def test_client_gen_with_new_format(self):
        """Test client generation with new format IDL."""
        idl_data = load_idl_if_exists("loopscale_v1.json")
        if idl_data:
            # This would test the client generation with new format
            # In a real test, we'd generate the client and verify the output
            assert True  # Placeholder for actual client gen test


class TestBackwardCompatibility:
    """Test that old code still works with the updates."""

    @pytest.mark.asyncio
    async def test_existing_code_with_old_idl(self):
        """Test that existing code patterns still work."""
        idl_data = load_idl_if_exists("kamino_lend_v4.json")
        if idl_data:
            provider = Provider.readonly()
            program_id = Pubkey.new_unique()

            # This is how existing code would create a program
            program = Program(idl_data, program_id, provider)

            # Verify everything still works
            assert program is not None
            assert program.program_id == program_id

            # Check that methods are available
            assert hasattr(program, "rpc")
            assert hasattr(program, "instruction")
            assert hasattr(program, "transaction")
            assert hasattr(program, "account")

    @pytest.mark.asyncio
    async def test_coder_backward_compatibility(self):
        """Test that Coder still works with old IDLs."""
        idl_data = load_idl_if_exists("adrena.json")
        if idl_data:
            from anchorpy3_core.idl import Idl

            idl = Idl.from_json(json.dumps(idl_data))
            coder = Coder(idl)

            # Verify all coder components work
            assert coder.accounts is not None
            assert coder.instruction is not None
            assert coder.events is not None
            assert coder.types is not None


from pathlib import Path

from anchorpy import Idl, Program
from solders.pubkey import Pubkey


def test_idls() -> None:
    idls = []
    programs = []
    for path in Path("tests/idls/").iterdir():
        raw = path.read_text()
        idl = Idl.from_json(raw)
        idls.append(idl)
        if "spl_token" not in str(path):
            program = Program(idl, Pubkey.default())
            programs.append(program)
    assert idls


def test_jet_enum() -> None:
    path = Path("tests/idls/jet.json")
    raw = path.read_text()
    idl = Idl.from_json(raw)
    program = Program(idl, Pubkey.default())
    expired_err = program.type["CacheInvalidError"].Expired
    assert expired_err(msg="hi").msg == "hi"


def test_switchboard_tuple() -> None:
    path = Path("tests/idls/switchboard.json")
    raw = path.read_text()
    idl = Idl.from_json(raw)
    program = Program(idl, Pubkey.default())  # noqa: F841


def test_clientgen_example() -> None:
    path = Path("tests/idls/clientgen_example_program.json")
    raw = path.read_text()
    Idl.from_json(raw)



from pathlib import Path

from anchorpy3_core.idl import Idl, IdlErrorCode
from autoflake import fix_code
from black import FileMode, format_str
from genpy import (
    Assign,
    Collection,
    FromImport,
    If,
    Import,
    Return,
    Statement,
    Suite,
)
from genpy import Function as UntypedFunction

from anchorpy.clientgen.common import _sanitize
from anchorpy.clientgen.genpy_extension import (
    Class,
    Function,
    InitMethod,
    IntDict,
    IntDictEntry,
    TypedParam,
    Union,
)
from anchorpy.error import LangErrorMessage, _LangErrorCode


def gen_from_code_fn(has_custom_errors: bool) -> Function:
    from_code_body = Return(
        "custom.from_code(code) if code >= 6000 else anchor.from_code(code)"
        if has_custom_errors
        else "anchor.from_code(code)"
    )
    from_code_return_type = (
        Union(["custom.CustomError", "anchor.AnchorError", "None"])
        if has_custom_errors
        else "typing.Optional[anchor.AnchorError]"
    )
    return Function(
        "from_code",
        [TypedParam("code", "int")],
        from_code_body,
        str(from_code_return_type),
    )


def gen_from_tx_error_fn(has_custom_errors: bool) -> Function:
    err_info_assign = Assign("err_info", "error.args[0]")
    err_code_assign = Assign("extracted", "extract_code_and_logs(err_info, PROGRAM_ID)")
    null_code_check = If("extracted is None", Return(None))
    final_return = Return("from_code(extracted[0])")
    fn_body = Suite(
        [
            err_info_assign,
            err_code_assign,
            null_code_check,
            final_return,
        ]
    )
    return_type = (
        "typing.Union[anchor.AnchorError, custom.CustomError, None]"
        if has_custom_errors
        else "typing.Optional[anchor.AnchorError]"
    )
    return Function(
        "from_tx_error",
        [TypedParam("error", "RPCException")],
        fn_body,
        return_type,
    )


def gen_custom_errors_code(errors: list[IdlErrorCode]) -> str:
    typing_import = Import("typing")
    error_import = FromImport(
        "anchorpy.error", ["ProgramError", "extract_code_and_logs"]
    )
    error_names: list[str] = []
    classes: list[Class] = []
    error_map_entries: list[IntDictEntry] = []
    for error in errors:
        code = error.code
        name = _sanitize(error.name)
        maybe_msg = error.msg
        msg = None if maybe_msg is None else f'"{maybe_msg}"'
        init_body = Statement(f"super().__init__({code}, {msg})")
        attrs = [
            InitMethod([], init_body),
            Assign("code", code),
            Assign("name", f'"{name}"'),
            Assign("msg", msg),
        ]
        klass = Class(name=name, bases=["ProgramError"], attributes=attrs)
        classes.append(klass)
        error_names.append(name)
        error_map_entries.append(IntDictEntry(code, f"{name}()"))
    type_alias = Assign("CustomError", Union(error_names))
    error_map = Assign(
        "CUSTOM_ERROR_MAP: dict[int, CustomError]", IntDict(error_map_entries)
    )
    from_code_body = Suite(
        [
            Assign("maybe_err", "CUSTOM_ERROR_MAP.get(code)"),
            If("maybe_err is None", Return("None")),
            Return("maybe_err"),
        ]
    )
    from_code_fn = Function(
        "from_code",
        [TypedParam("code", "int")],
        from_code_body,
        "typing.Optional[CustomError]",
    )
    return str(
        Collection(
            [typing_import, error_import, *classes, type_alias, error_map, from_code_fn]
        )
    )


def gen_custom_errors(idl: Idl, errors_dir: Path) -> None:
    errors = idl.errors
    if errors is None or not errors:
        return
    code = gen_custom_errors_code(errors)
    formatted = format_str(code, mode=FileMode())
    fixed = fix_code(formatted, remove_all_unused_imports=True)
    (errors_dir / "custom.py").with_suffix(".py").write_text(fixed)


def gen_anchor_errors_code() -> str:
    typing_import = Import("typing")
    error_import = FromImport("anchorpy.error", ["ProgramError"])
    error_names: list[str] = []
    classes: list[Class] = []
    error_map_entries: list[IntDictEntry] = []
    for variant in _LangErrorCode:
        name = _sanitize(variant.name)
        code = variant.value
        maybe_msg = LangErrorMessage.get(variant)
        msg = None if maybe_msg is None else f'"{maybe_msg}"'
        init_body = Statement(f"super().__init__({code}, {msg})")
        attrs = [
            UntypedFunction("__init__", ["self"], init_body),
            Assign("code", code),
            Assign("name", f'"{name}"'),
            Assign("msg", msg),
        ]
        klass = Class(name=name, bases=["ProgramError"], attributes=attrs)
        classes.append(klass)
        error_names.append(name)
        error_map_entries.append(IntDictEntry(code, f"{name}()"))
    type_alias = Assign("AnchorError", Union(error_names))
    error_map = Assign(
        "ANCHOR_ERROR_MAP: dict[int, AnchorError]",
        IntDict(error_map_entries),
    )
    from_code_body = Suite(
        [
            Assign("maybe_err", "ANCHOR_ERROR_MAP.get(code)"),
            If("maybe_err is None", Return("None")),
            Return("maybe_err"),
        ]
    )
    from_code_fn = Function(
        "from_code",
        [TypedParam("code", "int")],
        from_code_body,
        "typing.Optional[AnchorError]",
    )
    return str(
        Collection(
            [typing_import, error_import, *classes, type_alias, error_map, from_code_fn]
        )
    )


def gen_anchor_errors(errors_dir: Path) -> None:
    code = gen_anchor_errors_code()
    formatted = format_str(code, mode=FileMode())
    (errors_dir / "anchor").with_suffix(".py").write_text(formatted)


def gen_index_code(idl: Idl) -> str:
    has_custom_errors = bool(idl.errors)
    typing_import = Import("typing")
    rpc_exception_import = FromImport("solana.rpc.core", ["RPCException"])
    tx_status_import = FromImport(
        "solders.transaction_status",
        ["InstructionErrorCustom", "TransactionErrorInstructionError"],
    )
    preflight_error_import = FromImport(
        "solders.rpc.errors", ["SendTransactionPreflightFailureMessage"]
    )
    extract_code_and_logs_import = FromImport(
        "anchorpy.error", ["extract_code_and_logs"]
    )
    program_id_import = FromImport("..program_id", ["PROGRAM_ID"])
    anchor_import = FromImport(".", ["anchor"])
    re_import = Import("re")
    base_import_lines = [
        typing_import,
        re_import,
        tx_status_import,
        rpc_exception_import,
        preflight_error_import,
        extract_code_and_logs_import,
        program_id_import,
        anchor_import,
    ]
    custom_import_lines = [FromImport(".", ["custom"])] if has_custom_errors else []
    import_lines = base_import_lines + custom_import_lines
    from_code_fn = gen_from_code_fn(has_custom_errors)
    error_re_line = Assign(
        "error_re", r're.compile(r"Program (\w+) failed: custom program error: (\w+)")'
    )
    from_tx_error_fn = gen_from_tx_error_fn(has_custom_errors)
    return str(
        Collection(
            [
                *import_lines,
                from_code_fn,
                error_re_line,
                from_tx_error_fn,
            ]
        )
    )


def gen_index_file(idl: Idl, errors_dir: Path) -> None:
    code = gen_index_code(idl)
    path = errors_dir / "__init__.py"
    formatted = format_str(code, mode=FileMode())
    path.write_text(formatted)


def gen_errors(idl: Idl, root: Path) -> None:
    errors_dir = root / "errors"
    errors_dir.mkdir(exist_ok=True)
    gen_index_file(idl, errors_dir)
    gen_custom_errors(idl, errors_dir)
    gen_anchor_errors(errors_dir)



"""This module contains code for generating RPC functions."""
from typing import Any, Awaitable, Dict, Protocol

from anchorpy3_core.idl import IdlInstruction
from solana.rpc.commitment import Confirmed
from solana.rpc.core import RPCException
from solders.pubkey import Pubkey
from solders.signature import Signature

from anchorpy.error import ProgramError
from anchorpy.program.context import EMPTY_CONTEXT, Context, _check_args_length
from anchorpy.program.namespace.transaction import _TransactionFn
from anchorpy.provider import Provider


class _RpcFn(Protocol):
    """_RpcFn is a single RPC method generated from an IDL, sending a transaction paid for and signed by the configured provider."""  # noqa: E501

    def __call__(
        self,
        *args: Any,
        ctx: Context = EMPTY_CONTEXT,
    ) -> Awaitable[Signature]:
        """Call the function (this is just a protocol declaration).

        Args:
            *args: The positional arguments for the program. The type and number
                of these arguments depend on the program being used.
            ctx: non-argument parameters to pass to the method.
        """
        ...


def _build_rpc_item(  # ts: RpcFactory
    idl_ix: IdlInstruction,
    tx_fn: _TransactionFn,
    idl_errors: Dict[int, str],
    provider: Provider,
    program_id: Pubkey,
) -> _RpcFn:
    """Build the function that sends transactions for the given method.

    Args:
        idl_ix: The IDL instruction object.
        tx_fn: The function that generates the `Transaction` to send.
        idl_errors: Mapping of error code to error message.
        provider: Anchor Provider instance.
        program_id: The ID of the Anchor program.

    Returns:
        The RPC function.
    """

    async def rpc_fn(*args: Any, ctx: Context = EMPTY_CONTEXT) -> Signature:
        recent_blockhash = (
            await provider.connection.get_latest_blockhash(Confirmed)
        ).value.blockhash
        tx = tx_fn(
            *args, payer=provider.wallet.payer, blockhash=recent_blockhash, ctx=ctx
        )
        _check_args_length(idl_ix, args)
        try:
            return await provider.send(tx, ctx.options)
        except RPCException as e:
            err_info = e.args[0]
            translated_err = ProgramError.parse(err_info, idl_errors, program_id)
            if translated_err is not None:
                raise translated_err from e
            raise

    return rpc_fn



"""This module deals with generating transactions."""
from typing import Any, Protocol

from anchorpy3_core.idl import IdlInstruction
from solders.hash import Hash
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.message import Message
from solders.transaction import VersionedTransaction

from anchorpy.program.context import EMPTY_CONTEXT, Context, _check_args_length
from anchorpy.program.namespace.instruction import _InstructionFn


# ported from more-itertools
def _unique_everseen(iterable):
    """Yield unique elements, preserving order.

    >>> list(_unique_everseen('AAAABBBCCDAABBB'))
    ['A', 'B', 'C', 'D']
    """
    seenset = set()
    seenset_add = seenset.add
    seenlist = []
    seenlist_add = seenlist.append

    for element in iterable:
        try:
            if element not in seenset:
                seenset_add(element)
                yield element
        except TypeError:
            if element not in seenlist:
                seenlist_add(element)
                yield element


class _TransactionFn(Protocol):
    """A function to create a `Transaction` for a given program instruction."""

    def __call__(
        self, *args: Any, payer: Keypair, blockhash: Hash, ctx: Context = EMPTY_CONTEXT
    ) -> VersionedTransaction:
        """Make sure that the function looks like this.

        Args:
            *args: The positional arguments for the program. The type and number
                of these arguments depend on the program being used.
            payer: The transaction fee payer.
            blockhash: A recent blockhash.
            ctx: non-argument parameters to pass to the method.

        """
        ...


# ts TransactionNamespaceFactory.build
def _build_transaction_fn(
    idl_ix: IdlInstruction, ix_fn: _InstructionFn
) -> _TransactionFn:
    """Build the function that generates Transaction objects.

    Args:
        idl_ix: Instruction item from the IDL.
        ix_fn (_InstructionFn): The function that generates instructions.

    Returns:
        _TransactionFn: [description]
    """

    def tx_fn(
        *args: Any, payer: Keypair, blockhash: Hash, ctx: Context = EMPTY_CONTEXT
    ) -> VersionedTransaction:
        ixns: list[Instruction] = []
        _check_args_length(idl_ix, args)
        if ctx.pre_instructions:
            ixns.extend(ctx.pre_instructions)
        ixns.append(ix_fn(*args, ctx=ctx))
        if ctx.post_instructions:
            ixns.extend(ctx.post_instructions)
        ctx_signers = ctx.signers
        signers = [] if ctx_signers is None else ctx_signers
        all_signers = list(_unique_everseen([payer, *signers]))
        msg = Message.new_with_blockhash(ixns, payer.pubkey(), blockhash)
        return VersionedTransaction(msg, all_signers)

    return tx_fn



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class IncrementStateWhenPresentAccounts(typing.TypedDict):
    first_state: typing.Optional[Pubkey]
    second_state: Pubkey


def increment_state_when_present(
    accounts: IncrementStateWhenPresentAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["first_state"], is_signer=False, is_writable=True)
        if accounts["first_state"]
        else AccountMeta(pubkey=program_id, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["second_state"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf1!\xdd(\xa3Jy%"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT, CLOCK
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
from construct import Construct
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class InitializeWithValuesArgs(typing.TypedDict):
    bool_field: bool
    u8_field: int
    i8_field: int
    u16_field: int
    i16_field: int
    u32_field: int
    i32_field: int
    f32_field: float
    u64_field: int
    i64_field: int
    f64_field: float
    u128_field: int
    i128_field: int
    bytes_field: bytes
    string_field: str
    pubkey_field: Pubkey
    vec_field: list[int]
    vec_struct_field: list[types.foo_struct.FooStruct]
    option_field: typing.Optional[bool]
    option_struct_field: typing.Optional[types.foo_struct.FooStruct]
    struct_field: types.foo_struct.FooStruct
    array_field: list[bool]
    enum_field1: types.foo_enum.FooEnumKind
    enum_field2: types.foo_enum.FooEnumKind
    enum_field3: types.foo_enum.FooEnumKind
    enum_field4: types.foo_enum.FooEnumKind


layout = borsh.CStruct(
    "bool_field" / borsh.Bool,
    "u8_field" / borsh.U8,
    "i8_field" / borsh.I8,
    "u16_field" / borsh.U16,
    "i16_field" / borsh.I16,
    "u32_field" / borsh.U32,
    "i32_field" / borsh.I32,
    "f32_field" / borsh.F32,
    "u64_field" / borsh.U64,
    "i64_field" / borsh.I64,
    "f64_field" / borsh.F64,
    "u128_field" / borsh.U128,
    "i128_field" / borsh.I128,
    "bytes_field" / borsh.Bytes,
    "string_field" / borsh.String,
    "pubkey_field" / BorshPubkey,
    "vec_field" / borsh.Vec(typing.cast(Construct, borsh.U64)),
    "vec_struct_field"
    / borsh.Vec(typing.cast(Construct, types.foo_struct.FooStruct.layout)),
    "option_field" / borsh.Option(borsh.Bool),
    "option_struct_field" / borsh.Option(types.foo_struct.FooStruct.layout),
    "struct_field" / types.foo_struct.FooStruct.layout,
    "array_field" / borsh.Bool[3],
    "enum_field1" / types.foo_enum.layout,
    "enum_field2" / types.foo_enum.layout,
    "enum_field3" / types.foo_enum.layout,
    "enum_field4" / types.foo_enum.layout,
)


class InitializeWithValuesAccounts(typing.TypedDict):
    state: Pubkey
    payer: Pubkey


def initialize_with_values(
    args: InitializeWithValuesArgs,
    accounts: InitializeWithValuesAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=CLOCK, is_signer=False, is_writable=False),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdcI\x08\xd5\xb2E\xb5\x8d"
    encoded_args = layout.build(
        {
            "bool_field": args["bool_field"],
            "u8_field": args["u8_field"],
            "i8_field": args["i8_field"],
            "u16_field": args["u16_field"],
            "i16_field": args["i16_field"],
            "u32_field": args["u32_field"],
            "i32_field": args["i32_field"],
            "f32_field": args["f32_field"],
            "u64_field": args["u64_field"],
            "i64_field": args["i64_field"],
            "f64_field": args["f64_field"],
            "u128_field": args["u128_field"],
            "i128_field": args["i128_field"],
            "bytes_field": args["bytes_field"],
            "string_field": args["string_field"],
            "pubkey_field": args["pubkey_field"],
            "vec_field": args["vec_field"],
            "vec_struct_field": list(
                map(lambda item: item.to_encodable(), args["vec_struct_field"])
            ),
            "option_field": args["option_field"],
            "option_struct_field": (
                None
                if args["option_struct_field"] is None
                else args["option_struct_field"].to_encodable()
            ),
            "struct_field": args["struct_field"].to_encodable(),
            "array_field": args["array_field"],
            "enum_field1": args["enum_field1"].to_encodable(),
            "enum_field2": args["enum_field2"].to_encodable(),
            "enum_field3": args["enum_field3"].to_encodable(),
            "enum_field4": args["enum_field4"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



"""Contains code for parsing the IDL file."""
from typing import Sequence, TypedDict

import solders.pubkey
from anchorpy3_core.idl import IdlTypeDefinition
from borsh_construct import U8, CStruct, Vec

from anchorpy.borsh_extension import BorshPubkey


def _idl_address(program_id: solders.pubkey.Pubkey) -> solders.pubkey.Pubkey:
    """Deterministic IDL address as a function of the program id.

    Args:
        program_id: The program ID.

    Returns:
        The public key of the IDL.
    """
    base = solders.pubkey.Pubkey.find_program_address([], program_id)[0]
    return solders.pubkey.Pubkey.create_with_seed(base, "anchor:idl", program_id)


class IdlProgramAccount(TypedDict):
    """The on-chain account of the IDL."""

    authority: solders.pubkey.Pubkey
    data: bytes


IDL_ACCOUNT_LAYOUT = CStruct("authority" / BorshPubkey, "data" / Vec(U8))


def _decode_idl_account(data: bytes) -> IdlProgramAccount:
    """Decode on-chain IDL.

    Args:
        data: binary data from the account that stores the IDL.

    Returns:
        Decoded IDL.
    """
    return IDL_ACCOUNT_LAYOUT.parse(data)


TypeDefs = Sequence[IdlTypeDefinition]



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from .. import types


class GameJSON(typing.TypedDict):
    players: list[str]
    turn: int
    board: list[list[typing.Optional[types.sign.SignJSON]]]
    state: types.game_state.GameStateJSON


@dataclass
class Game:
    discriminator: typing.ClassVar = b"\x1bZ\xa6}Jdy\x12"
    layout: typing.ClassVar = borsh.CStruct(
        "players" / BorshPubkey[2],
        "turn" / borsh.U8,
        "board" / borsh.Option(types.sign.layout)[3][3],
        "state" / types.game_state.layout,
    )
    players: list[Pubkey]
    turn: int
    board: list[list[typing.Optional[types.sign.SignKind]]]
    state: types.game_state.GameStateKind

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Game"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Game"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Game"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Game":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Game.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            players=dec.players,
            turn=dec.turn,
            board=list(
                map(
                    lambda item: list(
                        map(
                            lambda item: (
                                None if item is None else types.sign.from_decoded(item)
                            ),
                            item,
                        )
                    ),
                    dec.board,
                )
            ),
            state=types.game_state.from_decoded(dec.state),
        )

    def to_json(self) -> GameJSON:
        return {
            "players": list(map(lambda item: str(item), self.players)),
            "turn": self.turn,
            "board": list(
                map(
                    lambda item: list(
                        map(
                            lambda item: (None if item is None else item.to_json()),
                            item,
                        )
                    ),
                    self.board,
                )
            ),
            "state": self.state.to_json(),
        }

    @classmethod
    def from_json(cls, obj: GameJSON) -> "Game":
        return cls(
            players=list(map(lambda item: Pubkey.from_string(item), obj["players"])),
            turn=obj["turn"],
            board=list(
                map(
                    lambda item: list(
                        map(
                            lambda item: (
                                None if item is None else types.sign.from_json(item)
                            ),
                            item,
                        )
                    ),
                    obj["board"],
                )
            ),
            state=types.game_state.from_json(obj["state"]),
        )



from pathlib import Path

from anchorpy import Idl
from anchorpy.program.namespace.instruction import _accounts_array
from pytest import mark
from solders.instruction import AccountMeta
from solders.keypair import Keypair


@mark.unit
def test_accounts_array() -> None:
    """Test accounts_array returns expected."""
    raw = Path("tests/idls/composite.json").read_text()
    idl = Idl.from_json(raw)
    dummy_a = Keypair()
    dummy_b = Keypair()
    comp_accounts = {
        "foo": {
            "dummy_a": dummy_a.pubkey(),
        },
        "bar": {
            "dummy_b": dummy_b.pubkey(),
        },
    }
    accounts_arg = idl.instructions[1].accounts
    acc_arr = _accounts_array(comp_accounts, accounts_arg)
    assert acc_arr == [
        AccountMeta(pubkey=dummy_a.pubkey(), is_signer=False, is_writable=True),
        AccountMeta(pubkey=dummy_b.pubkey(), is_signer=False, is_writable=True),
    ]
